"""
Metadata are all meta-information about a publication.
"""

from .model import SourceModel


class Metadata(SourceModel):
    """Dict-like object storing metadata of a publication

    Args:
        did (string): identifier for the publication
        trends:
        shared:
    """
    def __init__(self, data, document_id):
        super().__init__(data)
        self.did = document_id

    def __repr__(self):
        return '<Metadata.did={0.did}>'.format(self)



"""
Used to color some logs.
"""


class COLOR:
    """Settings for color in terminal"""
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'


def blue(string):
    """Convert a string into a string with blue font"""
    return '{}{}{}'.format(COLOR.OKBLUE, string, COLOR.ENDC)


def green(string):
    """Convert a string into a string with green font"""
    return '{}{}{}'.format(COLOR.OKGREEN, string, COLOR.ENDC)


def red(string):
    """Convert a string into a string with red font"""
    return '{}{}{}'.format(COLOR.FAIL, string, COLOR.ENDC)



"""
Publications are all documents which match the query you have defined in
your projects.
"""


from os import getcwd
from os.path import abspath
from reprlib import repr as trunc_repr

import requests

from .api import RadarlyApi
from .constants import PLATFORM, TONE
from .exceptions import PublicationUpdateFailed
from .metadata import Metadata
from .model import GeneratorModel, SourceModel
from .utils.misc import parse_image_url
from .utils.jsonparser import snake_dict
from .utils.checker import check_geocode, check_language, check_list


class Publication(SourceModel):
    """Object base on ``SourceModel`` storing information about the
    publication. The structure of the model can be drawn with the
    ``draw_structure`` method.

    Args:
        uid (str): unique identifier of the publication
        origin (dict): dictionary which contains information about the
            platform where the publication comes from.
        permalink (str): link to the publication
        lang (str): lang of the publication
        date (datetime.datetime): creation date of the publication
        impression (int): number of impressions on the publication
        reach (int): estimated number of people reached by the publication
        tone (str): tone of the publication
        category (str): category of the publications
        user (dict): information about the author of the publication
    """
    def __init__(self, data, project_id):
        super().__init__()
        self.pid = project_id
        data = snake_dict(data, blacklist=[['radar', 'tag']])
        super().add_data(data)

    def __repr__(self):
        try:
            publication_uid = trunc_repr(self['uid'])
        except KeyError:
            publication_uid = None
        return '<Publication.uid={}>'.format(publication_uid)

    @classmethod
    def fetch(cls, project_id, parameter, api=None):
        """
        Get publications stored inside a project.

        Args:
            project_id (int): identifier of a project
            parameter (SearchPublicationParameter): parameters object
                made with the SearchPublicationParameter instance, which will
                be used as payload data in POST request. See
                ``SearchPublicationParameter`` to know how to build this
                object.
            api (RadarlyApi, optional): API object used to perform request. If
                None, it will use the default API.
        Returns:
            list[Publication]:
        """
        api = api or RadarlyApi.get_default_api()
        url = api.router.publication['search'].format(project_id=project_id)
        data = api.post(url, data=parameter)
        return [
            Publication(item, project_id) for item in data['hits']
        ]

    @classmethod
    def fetch_all(cls, project_id, parameter, api=None):
        """Get all publications matching given parameters. It yields
        publications.

        Args:
            project_id (int): identifier of your project
            parameter (SearchPublicationParameter): parameters object
                made with the SearchPublicationParameter instance, which will
                be used as payload data in POST request. See
                ``SearchPublicationParameter`` to know how to build this
                object.
            api (RadarlyApi, optional): API object used to perform request. If
                None, it will use the default API.
        Returns:
            PublicationsGenerator: list of publications. On each iterations, a
            Publication is yielded until there is no more publication.
        """
        api = api or RadarlyApi.get_default_api()
        return PublicationsGenerator(parameter,
                                     project_id=project_id, api=api)

    def get_metadata(self, params=None, api=None):
        """This method allows users to get document’s metadata.

        Args:
            params (dict, optional): parameter sent in the GET request. Default
                to None.
        Returns:
            Metadata: object storing metadata information
        """
        api = api or RadarlyApi.get_default_api()
        url = api.router.publication['metadata'].format(project_id=self.pid)
        params = {} if params is None else params
        params.update(dict(
            platform=self['origin']['platform'],
            uid=self['uid'],
        ))
        res_data = api.get(url, params=params)
        return Metadata(res_data, self['uid'])

    def get_raw(self, params=None, api=None):
        """Get the raw content of the publication.

        Args:
            params (dict, optional): parameter sent in the GET request. Default
                to None.
        Returns:
            dict: dictionary storing the raw content of the publication
        """
        api = api or RadarlyApi.get_default_api()
        doc_platform = self['origin']['platform']
        available_platform = [
            PLATFORM.FORUM,
            PLATFORM.BLOG,
        ]
        assert doc_platform in available_platform, \
            "{} is not compatible with raw content".format(doc_platform)
        url = api.router.publication['raw'].format(project_id=self.pid)
        params = {} if params is None else params
        params.update(dict(
            platform=doc_platform,
            uid=self['uid'],
        ))
        res_data = api.get(url, params=params)
        return res_data

    def set_tags(self, *args, **kwargs):
        """Update some information about a publication in Radarly. This
        method is based on the ``set_publication_tags`` function defined in
        the same module.

        .. warning:: Unlike the ``set_publication_tags``, this function returns
            None in case of success (the ``set_publication_tags`` returns a
            ``Publication`` object).

        Args:
            tone (str): tone of the publication. Can be `positive`,
                `negative`, `mixed` or `neutral`.
            language (str): alpha-2, alpha-3, or name of the language
            country (str): alpha-2, alpha-3 or name of the country
            keywords (list[str]): list of keywords for the publication
            custom_tags (dict[str -> list[str]]): value of the custom tags
                to set. The template for this argument is::

                {<label of the custom_tag>: [<label of the subtag>]}.

                Example: Given two tags (the first one named ``Product``
                with ``Shoes``, ``T-Shirt`` and ``Clothes`` as subtags and
                the second one named ``Price`` with ``High``, ``Medium``
                and ``Low`` as subtags), a valid value for the ``custom_tags``
                could be::

                    {'Product': ['Clothes', 'T-Shirt'], 'Price': ['High']}
        Raises:
            PublicationUpdateFailed: error raised if the publication failed
        Returns:
            None
        """

        publication = set_publication_tags(
            *args,
            project_id=getattr(self, 'pid'),
            uid=getattr(self, 'uid'),
            platform=getattr(self, 'origin')['platform'],
            **kwargs
        )
        self.add_data(publication.__dict__)
        return None


    def download(self, output_dir=None, chunk_size=1024):
        """Download the publication if it is an image or video.

        .. warning:: This function will not raised an error even if the
            download fails. To know if all the download succeed, compare the
            media object of a publication with the response of the function.

        Args:
            output_dir (str, optional): folder where the downloaded images must
                be saved. The folder must already exists. Default to the
                current working directory.
            chunk_size (int, optional): chunk size used during the file
                download with ``requests``. Default to 1024.
        Returns:
            dict[str]: filepath of the downloaded medias. This dictionary has
            quite the same structure of the ``media`` attribute of the
            publication.
        """
        def download_content(content_link, output_dir):
            """Download the content of a media and save it in a existing
            directory.

            Args:
                content_link (str):
                output_dir (str):
            Returns:
                dict: local version of the media object
            """
            if content_link is None: return None
            res = requests.get(content_link, stream=True)
            try:
                res.raise_for_status()
            except requests.exceptions.HTTPError:
                return None
            img_name, img_format = parse_image_url(res.url)
            filepath = '{}/{}.{}'.format(output_dir, img_name, img_format)

            with open(filepath, mode='wb') as image_file:
                for chunk in res.iter_content(chunk_size=chunk_size):
                    image_file.write(chunk)

            return abspath(filepath)

        output_dir = output_dir or getcwd()

        media_links = dict(
            image=[],
            video=[]
        )
        if self['media'] and self['media']['image']:
            downloaded_images = [
                download_content(item, output_dir) for item in self['media']['image']
            ]
            media_links['image'].extend(list(filter(None, downloaded_images)))
        if self['media'] and self['media']['video']:
            downloaded_videos = [
                {
                    'url': download_content(item['url'], output_dir),
                    'thumbnail': download_content(item['thumbnail'], output_dir)
                } for item in self['media']['video']
            ]
            media_links['video'].extend(
                filter(lambda x: x['url'] and x['thumbnail'], downloaded_videos)
            )

        return media_links


class PublicationsGenerator(GeneratorModel):
    """Generator which yields all publications matching some payload.

    Args:
        search_param (SearchPublicationParameter):
        project_id (int): identifier of the project
        api (RadarlyApi): api to use to perform requests
    Yields:
        Publication:
    """
    def _fetch_items(self):
        """Get next range of publications"""
        url = self._api.router.publication['search'].format(
            project_id=self.project_id
        )
        res_data = self._api.post(url, data=self.search_param)
        self.total = res_data['total']
        self._items = (
            Publication(item, self.project_id)
            for item in res_data['hits']
        )
        div = self.total // self.search_param['limit']
        reste = self.total % self.search_param['limit']
        self.total_page = div
        if reste != 0: self.total_page += 1
        self.search_param = self.search_param.next_page()

    def __repr__(self):
        return '<PublicationsGenerator.total={}.total_page={}>'.format(
            self.total, self.total_page
        )


def set_publication_tags(project_id, uid, platform,
                         tone=None, language=None, country=None,
                         keyword=None, custom_tags=None, api=None):
    """Update some fields in a publication in Radarly. Only specific field
    can be updated. This function is outside the ``Publication`` class in order
    to give you the opportunity to update a publication without having the full
    object. The ``set_tags`` method of ``Publication`` uses this function so
    there is no big differences between theses two functions.

    Args:
        tone (str): tone of the publication. Can be `positive`,
            `negative`, `mixed` or `neutral`.
        language (str): alpha-2, alpha-3, or name of the language
        country (str): alpha-2, alpha-3 or name of the country
        keywords (list[str]): list of keywords for the publication
        custom_tags (dict[str -> list[str]]): value of the custom tags
            to set. The tags must already exist. The template for this
            argument is::

            {<label of the custom_tag>: [<label of the subtag>]}.

            Example: Given two tags (the first one named ``Product``
            with ``Shoes``, ``T-Shirt`` and ``Clothes`` as subtags and
            the second one named ``Price`` with ``High``, ``Medium``
            and ``Low`` as subtags), a valid value for the ``custom_tags``
            could be::

                {'Product': ['Clothes', 'T-Shirt'], 'Price': ['High']}
    Raises:
        PublicationUpdateFailed: error raised if the publication failed
    Returns:
        Publication: publication which was updated
    """

    def check_update(pub, tone, language, country,
                     keyword, custom_tags):
        """Check the publication's update."""
        not_updated_fields = []
        if tone and pub['tone'] != tone:
            not_updated_fields.append('tone')
        if keyword and pub['keyword'] != keyword:
            not_updated_fields.append('keyword')
        if language and pub['lang'] != language:
            not_updated_fields.append('lang')
        if country and pub['geo']['inferred']['country'] != country:
            not_updated_fields.append('geo.inferred.country')
        if custom_tags:
            _ = [
                not_updated_fields.append('radar.tag.custom.{}'.format(key))
                for key in custom_tags
                if pub['radar']['tag']['custom'][key] != custom_tags[key]
            ]
        if not_updated_fields:
            raise PublicationUpdateFailed(fields=not_updated_fields)
        return None

    payload = {}
    if tone:
        _ = TONE.check(tone)
        payload['tone'] = tone
    if language:
        language = check_language(language)[0].lower()
        payload['lang'] = language
    if keyword is not None:
        check_list(keyword, str, "`keyword` must be a list of string.")
        payload['keyword'] = {'set': keyword}
    if country:
        payload['country'] = check_geocode(country)
    if custom_tags:
        _ = {
            key: check_list(custom_tags[key], str,
                            ("The value for the field `{}` must be a "
                             "list of `{}` element").format(key, str(str)))
            for key in custom_tags
        }
        custom_tags_setter = {
            label: {'set': custom_tags[label]} for label in custom_tags
        }
        payload['radar'] = {'tag': {'custom': custom_tags_setter}}

    params = dict(
        uid=uid,
        platform=platform,
    )
    payload = {
        "doc": payload
    }
    api = api or RadarlyApi.get_default_api()
    url = api.router.publication['set_tag'].format(
        project_id=project_id
    )
    publication = api.post(url, params=params, data=payload)
    publication = Publication(data=publication, project_id=project_id)

    check_update(publication, tone, language, country,
                 keyword, custom_tags)

    return publication



"""
Some functions and classes used internally.
"""

import re
from lxml import html

from .misc import flat


def instance_builder(cls, data, *args, **kwargs):
    """Build an object of a specific class. This function is scalable
    (if a list is set as argument, return a list of object; else just
    one object)
    """
    if not data:
        return data
    elif isinstance(data, dict):
        return cls(data, *args, **kwargs)
    elif isinstance(data, list):
        return [cls(item, *args, **kwargs) for item in data]
    raise TypeError


class CallableDict(dict):
    """Dict which can be called (proxy for __getitem__ method). This object has three
    additional features:
    * it converts all the keys of data source into string
    * the key before a search will be converted into a string
    * if the value asked is not found, the value will be returned
    """
    def __init__(self, *datas):
        super().__init__()
        for data in datas:
            data = dict([(str(key), data[key]) for key in data])
            self.update(data)

    def __missing__(self, key):
        return str(key)

    def __getitem__(self, key):
        key = str(key).replace('focus_', '')
        return super().__getitem__(key)

    def __call__(self, key):
        return self[key]


def id_to_value(data, dtype):
    """Based on some data, build a dictionary to translate an ID to a
    a label.

    Args:
        data (list[object]):
        dtype (str):
    Raises:
        ValueError: error raised if dtype is not 'focuses' or 'tags'
    Returns:
        dict: dictionary where the key are ID of the object and the value,
        the label.
    """
    if dtype == 'focuses':
        return dict([
            (str(focus['id']), focus['label']) for focus in data
        ])
    elif dtype == 'tags':
        return dict(flat([[
            (str(subtag['id']), subtag['value']) for subtag in tag.subtags
        ] for tag in data]))
    raise ValueError


def parse_struct_stat(stats):
    """Parse the structure of stats values in influencer data in order
    to return a pandas-compatible object.
    """
    data = dict()
    for stat in stats:
        for item in stats[stat]:
            for metric in item['counts']:
                data.setdefault(metric, dict())
                data[metric].update({
                    (stat, item['term']): item['counts'][metric]
                })
    return data


def parse_struct_interval(intervals):
    """Parse the response to get distribution which can be easily converted
    with pandas"""
    data = dict()
    for item in intervals:
        for metric in item['counts']:
            data.setdefault(metric, {})
            data[metric].update({item['date']: item['counts'][metric]})
    return data


def _parse_error_response(response):
    """Parse an error response made with the request module
    in order to extract information about the error.

    Args:
        response (requests.Response): error response get from
            a request made with requests.
    Returns:
        dict: dictionary with information about the error,
        parsed from the content of th response
    """
    error_data = dict()
    if response.ok:
        return error_data
    error_data['error_code'] = response.status_code
    content_type = response.headers.get('Content-Type', '')
    if content_type == 'text/html' or '<!DOCTYPE html>' in response.text:
        document = html.fromstring(response.text)
        error_data['error_type'] = document.xpath('//title/text()')[0]
        try:
            element = document.xpath("//body")[0]
            message = element.text_content()
        except IndexError:
            message = ''
        finally:
            message = message.replace('\n', ' ')
            message = re.sub(r'[ ]+', ' ', message).strip()
            error_data['error_message'] = message
    elif content_type == 'application/json':
        error_data.update(response.json())
    return error_data



"""
The Dashboard is an object in order to organize your project. It is a set
of focuses, social accounts and benchmark_entities.

In the Python client, a ``Dashboard`` object stored the ID of all the objects
contained in the dashboard.
"""

from .model import SourceModel
from .utils._internal import instance_builder


class Dashboard(SourceModel):
    """
    Object storing ID's of the focuses, social_accounts and benchmark entities
    which are grouped in the dashboard. The full objects are stored in the Project
    object.

    Args:
        id (int): unique identifier for the dashboard
        label (str): name of the dashboard (it correponds to the name in
            Radarly)
        description (None or str): if available, describes the purpose of
            the dashboard
        created (datetime): creation date of the dashboard
        focuses (list[int]): focuses' ids of the query in the dashboard
        social_accounts (list[dict]): list of all social accounts
            recorded in the dashboard. They are grouped by platforms.
        benchmark_entities (list[dict]): list of all benchmark entities
            in the dashboard.
        ...
    """
    def __init__(self, data):
        super().__init__()
        translator = dict(
            focuses=lambda focuses: [focus['id'] for focus in focuses],
        )
        super().add_data(data, translator)

    @classmethod
    def _builder(cls, dashboards):
        return instance_builder(cls, dashboards)

    def __repr__(self):
        dashboard_id, label = getattr(self, 'id'), getattr(self, 'label')
        return '<Dashboard.id={}.label={}>'.format(dashboard_id, label)



"""
Hooks used to parse JSON data contained in the response of a request.
"""

import re
from datetime import datetime
import json

import pytz
from dateutil.parser import parse

from .misc import to_snake_case


_BLACKLIST_PATH = [
    ['hits', 'radar', 'tag'],
    ['radar', 'tag'],
    ['dots', 'stats'],
]


def decode_value(value, key=None):
    """Try to convert a string into a specific Python object"""
    pattern_date = re.compile(
        r'^ *\d{4}-\d{2}-\d{2}T\d{2}:\d{2}:\d{2}.\d{3}Z *$'
    )

    if isinstance(value, str) and pattern_date.match(value):
        return parse(value.strip(), ignoretz=True)
    elif key == 'timezone' and value in pytz.all_timezones_set:
        return pytz.timezone(value)

    return value


def radarly_decoder(obj):
    """Hook which can be used in the `dumps` function of `json` module."""
    return dict(
        (to_snake_case(key), decode_value(obj[key], key)) for key in obj
    )


class RadarlyEncoder(json.JSONEncoder):
    """Encoder to serialize an object which inherits from SourceModel"""
    def default(self, obj):
        if 'radarly' in str(type(obj)):
            return dict(
                (key, dict(obj)[key]) for key in obj.keys()
                if not key.startswith('_') and not callable(dict(obj)[key])
            )
        if 'pytz.' in str(type(obj)):
            return obj.zone
        elif isinstance(obj, datetime):
            return obj.strftime('%Y-%m-%dT%H:%M:%SZ')
        return json.JSONEncoder.default(self, obj)


def snake_dict(data, blacklist=None, path=[]):
    """Convert all the keys of a dictionary into snake case format. It will
    also convert some values into Python object (like all the dates or timezone
    field).
    Map automatically if the element is a list. Some path can be blacklisted
    and so they will not be converted in to snake_case format.

    Args:
        data (object): dictionary or list to convert
        blacklist (list[list[str]]): path to ignored during the parsing.
            Example: ``[['radar', 'tag', 'custom']]``
        path (list[str]): internal use.
    Returns:
        object: the same object with all keys converted into snake cas format
            and some values converted into Python object.
    """
    if isinstance(data, list):
        return [snake_dict(item,
                           blacklist=blacklist,
                           path=path) for item in data]

    if isinstance(data, dict) and path not in blacklist:
        return {
            to_snake_case(key): snake_dict(data[key],
                                           path=path + [key],
                                           blacklist=blacklist)
            for key in data
        }

    return decode_value(data, path[-1])



"""
The API is submitted to several rates limits. The rates are returned in the
headers of each request. This modules defines the object used to parse the
request's headers and store this rates.
"""

import copy
import re

from .utils.router import Router


class RateConf:
    """Object storing information about the rate limit on each url of the
    ``Router`` object"""
    slow = [
        Router.influencer['search'],
        Router.distribution['fetch'],
        Router.analytics['global'],
        Router.cloud['fetch'],
        Router.localization['fetch'],
    ]
    medium = [
        Router.publication['search'],
        Router.publication['metadata'],
        Router.publication['raw'],
    ]

    @classmethod
    def get_category(cls, url):
        """Get the rate limit category of a URL.

        Args:
            url (string):
        Returns:
            string: category for this URL ('slow', 'medium' or 'default')
        """
        def into_pattern(path_url):
            """Transform a string with format option into a compiled regular
            expression"""
            if isinstance(path_url, str):
                format_pattern = r'{[a-zA-Z0-9_]*}'
                path_url = re.sub(format_pattern, '[a-zA-Z0-9]*', path_url)
                return re.compile(path_url)
            elif isinstance(path_url, (list, tuple)):
                return [into_pattern(url) for url in path_url]
            raise TypeError("'path_url' must be a string or a list")
        match_slow = any(
            [item.search(url) for item in into_pattern(cls.slow)]
        )
        match_medium = any(
            [item.search(url) for item in into_pattern(cls.medium)]
        )
        if match_slow:
            return 'slow'
        elif match_medium:
            return 'medium'
        return 'default'


class RateLimit:
    """Object which will count the remaining request"""
    def __init__(self):
        default = dict(limit=0, remaining=5000, reset=0)
        self.slow = copy.deepcopy(default)
        self.medium = copy.deepcopy(default)
        self.default = copy.deepcopy(default)

    def __repr__(self):
        log = []
        for category in ['slow', 'medium', 'default']:
            limit = getattr(self, category)
            percent = 100 * limit['limit'] / (limit['limit'] +  limit['remaining'])
            log.append((category, percent))
        log = '.'.join(['{}:{:.2f}%'.format(*item) for item in log])
        return '<RateLimit.{}>'.format(log)

    def update(self, url, data):
        """Parse a dictionnary which corresponds to the header of a response
        in order to get the current limit. The information retrieved is
        stored in the object.

        Args:
            url (string): url of the request
            data (dict): headers of the requests
        Returns:
            None:
        """
        right_limit = getattr(self, RateConf.get_category(url))
        right_limit['limit'] = int(data.get('X-Rate-Limit-Limit', right_limit['limit']))
        right_limit['remaining'] = int(
            data.get('X-Rate-Limit-Remaining', right_limit['remaining'])
        )
        right_limit['reset'] = int(data.get('X-Rate-Limit-Reset', right_limit['reset']))
        return None

    def is_reached(self, url):
        """Whether or not the current user can run some requests.

        Args:
            url (string): url which will be fectch by the pending request
        """
        right_limit = getattr(self, RateConf.get_category(url))
        return right_limit['remaining'] <= 0



"""
Module to handle all roads used to retrieve data.
"""


class Router:
    """Roads defined by the Radarly API. Each attribute is a dictionary
    storing URLs of the same category."""
    analytics = {
        'global': '/projects/{project_id}/insights.json',
        'occupation': '/projects/{project_id}/insights/occupation.json'
    }
    benchmark = {
        'fetch': '/projects/{project_id}/benchmark.json',
    }
    cloud = {
        'fetch': '/projects/{project_id}/insights/cloud.json',
    }
    cluster = {
        'fetch': '/projects/{project_id}/stories.json',
    }
    corpora = {
        'fetch_media': '/projects/{project_id}/corpora/{corpora_id}.json',
    }
    distribution = {
        'fetch': '/projects/{project_id}/inbox/distribution.json',
    }
    geogrid = {
        'fetch': '/projects/{project_id}/geogrid.json',
    }
    host = {
        'prod': 'https://radarly.linkfluence.com',
    }
    influencer = {
        'search': '/projects/{project_id}/influencers.json',
        'find': '/projects/{project_id}/influencer.json',
    }
    localization = {
        'fetch': '/projects/{project_id}/insights/geo/{region_type}.json',
    }
    oauth = {
        'prod': 'https://oauth.linkfluence.com/oauth2/token',
    }
    pivot_table = {
        'fetch': '/projects/{project_id}/insights/pivot.json',
    }
    project = {
        'find': '/projects/{project_id}.json',
    }
    publication = {
        'search': '/projects/{project_id}/inbox/search.json',
        'metadata': '/projects/{project_id}/documents.json',
        'raw': '/projects/{project_id}/documents/raw.json',
        'set_tag': '/projects/{project_id}/documents.json',
    }
    social_performance = {
        'fetch': '/projects/{project_id}/performance.json',
    }
    topicwheel = {
        'fetch': '/projects/{project_id}/topicwheel.json',
    }
    user = {
        'me': '/users.json',
    }



"""
This module is the core module of the Python client. It handles all the
requests made to the API and parses the response in order make the
interactions with the API as easy as possible. The initialization of the
RadarlyApi's object defined here is the first step to start using the API.
We advise you to use a default RadarlyAPI object otherwise you must specify
a RadarlyApi object each time you want to retrieve data from the API.
To initialize a default API, simply run:

>>> from radarly import RadarlyApi
>>> RadarlyApi.init(client_id=<client_id>, client_secret=<client_secret>)
"""

import json
from datetime import datetime
from os import getenv

import requests

from .auth import RadarlyAuth
from .exceptions import (AuthenticationError, NoInitializedApi,
                         RadarlyHTTPError, RateReached)
from .rate import RateLimit
from .utils._internal import _parse_error_response
from .utils.jsonparser import snake_dict as _decoder, _BLACKLIST_PATH
from .utils.router import Router

__all__ = ['RadarlyApi']


class RadarlyApi: # pylint: disable=R0902
    """Main interface with the Radarly API. It defines several methods in
    order to ease the interactions with the Radarly API. For example, it
    defines the methods in order to authenticate to the API or to refresh the
    tokens. Thanks to the current implementation of the package, you can use
    ``radarly`` without using any of the ``RadarlyApi``'s methods.

    Args:
        client_id (str): client ID which is given by Linkfluence in order to
            use the API. Can be set at initialization. If this argument is
            set to None during initialization, it will try to recover the
            RADARLY_CLIENT_ID environment variable.
        client_secret (str): code given by Linkfluence in order to ensure that
            the right user is using the API. Can be set at initialization.
            If this argument is set to None during initialization, it will
            try to recover the RADARLY_CLIENT_SECRET environment variable.
        scope (list[str]): scopes of the API. If None, the default value will
            be used (which is ``['listening', 'social-performance']``). Can be
            set at initialization.
        version (str): version of the API you want to use. Can be set at
            initialization. Default to ``1.1``.
        access_token (str): code generated by the API in order to authenticate
            each request. This code is automatically retrieved by the
            RadarlyApi during its initilization. The access token has a limited
            lifetime and must be refreshed when it has expired.
        refresh_token (str): code used to generate a new access token when
            you've got an expired token. The RadarlyApi can automatically use
            this refresh token to update the access token when it's expired.
        autorefresh (bool, optional): Whether or not refresh the access token
            when it's expired. Can be set during initialization. Default to True.
        timeout (float): maximum duration during which the API will wait for an
            answer from the Radarly's server. Can be set at initialization.
        proxies (dict): proxies used for the requests. The proxies must be set
            in the same way that in the ``requests`` module. Can be set at
            initialization.
        rates (RateLimit): RateLimit object which can be used to know the
            current number of requests made and how many left you can do.
    """
    _default_api = None

    def __init__(self, # pylint: disable=C0303
                 client_id=None,
                 client_secret=None,
                 scope=None,
                 autorefresh=True,
                 timeout=None,
                 proxies=None,
                 version='1.1',
                 router=None,
                 environment='prod',
                 authenticate=True):
        client_id = client_id or getenv('RADARLY_CLIENT_ID')
        client_secret = client_secret or getenv('RADARLY_CLIENT_SECRET')
        if not(client_id and client_secret):
            raise KeyError(("Neither client_id nor client_secret has been set "
                            "during the initialization of the API. Furthermore "
                            "RADARLY_CLIENT_ID and RADARLY_CLIENT_SECRET "
                            "variables have not be found in environment "
                            "variables. You must either specify the client_id "
                            "and client_secret arguments or set the right "
                            "environment variables to start using the "
                            "client."))
        self.client_id = client_id
        self.client_secret = client_secret
        self.access_token = None
        self.refresh_token = None
        self.timeout = timeout
        self.proxies = proxies
        self.autorefresh = autorefresh
        self.last_refresh = datetime.now()
        self._auth = None
        self.rates = RateLimit()
        self.scope = scope or [
            'listening',
            'social-performance',
        ]
        self.version = version
        self.environment = environment
        self.router = router or \
            type('Router', Router.__bases__, dict(Router.__dict__))
        if authenticate:
            self.authenticate()

    def __repr__(self):
        return '<RadarlyAPI.client_id={.client_id}>'.format(self)

    @classmethod
    def init(cls, *args, **kwargs):
        """
        Initialize a RadarlyAPI and set it as default api. The default API is
        the one that will be used each time you made a request if no API is
        passed in the parameter.

        Args:
            client_id (str): client_id you want to use to access data
            client_secret (str): secret token given by Linkfluence.
            autorefresh (bool, optional): whether or not refresh automatically
                the acces_token when it's expired.
            scope (list[str]): list of scope
            timeout (float): timeout for the requests
            proxies (dict): proxies to use for the requests.
        """
        api = cls(*args, **kwargs)
        cls.set_default_api(api)
        return api

    @classmethod
    def get_default_api(cls):
        """
        Retrieve the default api previously set by the user.

        Raises:
            AssertionError: raised if no api was set as default api.
                Use ``init`` method to set a default API.
        Returns:
            RadarlyApi: api object you have previously initialized
        """
        if cls._default_api is None:
            raise NoInitializedApi
        return cls._default_api

    @classmethod
    def set_default_api(cls, api):
        """Set a default API for the client. This method is automatically
        called when you use the ``init`` class method."""
        assert isinstance(api, cls), ("Only a RadarlyApi object can "
                                      "be set as default api")
        cls._default_api = api
        return None

    def request(self, verb, url, **kwargs):
        """
        Send a request using the request module. Some pre- and post-tasks
        are computed each time in order to actualize the rates information
        and check whether or not the request is a success. This method uses
        the same parameters as request function of requests module so you can
        easily made your own authenticated request.

        Args:
            verb (string): method used for the request
            url (string): url to ask
            **kwargs: keywords arguments sent with request
        Raises:
            HTTP Error: raised if the request failed for an unknown cause
        Returns:
            dict: corresponds to the response data of the answer
        """
        root_url = self.router.host[self.environment]
        url = url.strip('/')
        if self._auth is None:
            self.authenticate()
        if self.version and not url.startswith(self.version) \
            and not url.startswith(root_url):
            url = '{}/{}'.format(self.version, url)
        url = url if root_url in url  \
            else '{}/{}'.format(root_url, url)
        kwargs.setdefault('headers', {})
        kwargs['headers'].setdefault('Content-Type', 'application/json')
        if ('data' in kwargs and
                kwargs['headers']['Content-Type'] == 'application/json'):
            kwargs['data'] = json.dumps(kwargs['data'])
        kwargs.setdefault('timeout', self.timeout)
        kwargs.setdefault('proxies', self.proxies)
        if self.rates.is_reached(url):
            raise RateReached('No more request available')

        res = requests.request(verb, url, auth=self._auth, **kwargs)

        error_data = _parse_error_response(res)
        error_type = error_data.get('error_type', '')
        if self.autorefresh and error_type == 'ExpiredTokenException':
            self.refresh()
            res = requests.request(verb, url, auth=self._auth, **kwargs)
        if not res.ok:
            raise RadarlyHTTPError(response=res)

        self.rates.update(url, res.headers)

        return _decoder(res.json(), blacklist=_BLACKLIST_PATH)

    def get(self, url, **kwargs):
        """Shortcut for the ``request`` method with 'GET' as verb.

        Args:
            url (string): url to ask
            **kwargs: keywords arguments sent with ``request``
        Raises:
            HTTP Error: raised if the request failed for an unknown cause
        Returns:
            dict: corresponds to the response data of the answer
        """
        return self.request('GET', url, **kwargs)

    def post(self, url, **kwargs):
        """Shortcut for the ``request`` method with 'POST' as verb.

        Args:
            url (string): url to ask
            **kwargs: keywords arguments sent with request
        Raises:
            HTTP Error: raised if the request failed for an unknown cause
        Returns:
            dict: corresponds to the response data of the answer
        """
        return self.request('POST', url, **kwargs)

    def put(self, url, **kwargs):
        """Shortcut for the ``request`` method with 'PUT' as verb.

        Args:
            url (string): url to ask
            **kwargs: keywords arguments sent with request
        Raises:
            HTTP Error: raised if the request failed for an unknown cause
        Returns:
            dict: corresponds to the response data of the answer
        """
        return self.request('PUT', url, **kwargs)

    def authenticate(self):
        """
        Enable the authentification to Radarly API. The Radarly API is
        currently using the OAUTH2 system for authorization : the
        (client_id, client_secret) will be used to generate an access_token
        and a refresh_token.

        Raises:
            BadAuthentication: raised if client_id or client_secret is
                incorrect
        Returns:
            None:
        """
        data = dict(
            client_id=self.client_id,
            client_secret=self.client_secret,
            grant_type='client_credentials',
            scope=' '.join(self.scope)
        )
        headers = {'Content-Type': 'application/x-www-form-urlencoded'}
        kwargs = dict(
            data=data,
            headers=headers,
            proxies=self.proxies,
            timeout=self.timeout
        )
        url = self.router.oauth[self.environment]
        auth_response = requests.request('POST', url,
                                         **kwargs)
        try:
            auth_response.raise_for_status()
        except requests.exceptions.HTTPError:
            auth_response = auth_response.json()
            raise AuthenticationError(auth_response.get('error'))
        auth_response = auth_response.json()
        self.scope = auth_response.get('scope', '').split(' ')
        self.access_token = auth_response.get('access_token')
        self.refresh_token = auth_response.get('refresh_token')
        self._auth = RadarlyAuth(self.access_token)
        return None

    def refresh(self):
        """
        Refresh the access_token using the refresh_token as soon as the access
        token has expired. The auto-refresh behaviour can be ignored with
        the autorefresh attribute.
        """
        data = dict(
            client_id=self.client_id,
            client_secret=self.client_secret,
            grant_type='refresh_token',
            refresh_token=self.refresh_token,
        )
        headers = {
            'Content-Type': 'application/x-www-form-urlencoded'
        }
        url = self.router.oauth[self.environment]
        auth_response = requests.request('POST', url,
                                         data=data, headers=headers,
                                         proxies=self.proxies, timeout=self.timeout)
        auth_response = auth_response.json()
        self.access_token = auth_response.get('access_token')
        self.refresh_token = auth_response.get('refresh_token')
        self._auth = RadarlyAuth(self.access_token)
        self.last_refresh = datetime.now()
        return None



"""
.. _API call: https://api.linkfluence.com/officialdoc/architecture/project.html#get-project-infos

As in Radarly, the ``Project`` object contains all the needed data
in order to start exploring a project. The ``Project`` object is built
dynamically using the response got from the `API call`_ in order to retrieve
information about a project. Several methods have been defined in order
to, for example, retrieve publications inside the project, know the
influencers or get some statitics about publications inside your project.
"""

from .analytics import Analytics
from .api import RadarlyApi
from .benchmark import Benchmark
from .cloud import Cloud
from .cluster import Cluster
from .corpus import InfoCorpus
from .dashboard import Dashboard
from .distribution import Distribution
from .focus import Focus
from .geogrid import GeoGrid
from .influencer import Influencer
from .localization import Localization
from .milestone import Milestone
from .model import SourceModel
from .pivottable import PivotTable
from .publication import Publication
from .socialaccount import SocialAccount
from .socialperformance import SocialPerformance
from .tag import Tag
from .topic import Entities, TopicWheel
from .utils._internal import id_to_value, instance_builder


class Project(SourceModel):
    """Object to explore a project in Radarly. This object is made
    dynamically but you can explore it using the methods defined in
    ``SourceModel``.

    Example:

        >>> project.keys()
        {'id', 'account_id', 'label', ..., 'brand_logos'}
        >>> project.draw_structure(max_depth=1)
        Project (Project)
         | id (str)
         | doc_count (int)
         ...
        >>> project['tags']
        [<Tag.id=xxxx.type=user>, <Tag.id=xxxx.type=custom>, ...]
        >>> project['$.tags(type=custom).subtags']
        ...

    Here is information about some useful attributes of the ``Project``
    object.

    Args:
        id (string): the unique ID used in our database to identify the project
        docs_count (int): the actual number of documents inside your project
        created (datetime.datetime): creation date of the project
        updated (datetime.datetime): last date the project was updated
        focuses (Focuses): list of all focuses (queries) of the project
        dashboards (Dashboards): list of all dashboards created to organize your
            project.
        tags (Tags): list of all tags of the project
        social_accounts (dict): information about the social accounts
            registred in your project.
        benchmark_entities (list[dict]): information about the accounts
            available for a benchmark
    """

    def __init__(self, data, translator=None, api=None):
        base_translator = dict(
            tags=Tag._builder,
            dashboards=Dashboard._builder,
            focuses=Focus._builder,
            milestones=Milestone._builder,
            social_accounts=SocialAccount._builder,
            corpora=InfoCorpus._builder
        )
        translator = translator or dict()
        base_translator.update(translator)
        super().__init__(data, base_translator)
        api = api or RadarlyApi.get_default_api()

    def __repr__(self):
        pid, label = getattr(self, 'id'), getattr(self, 'label')
        return '<Project.pid={}.label={}>'.format(pid, label)

    @classmethod
    def _builder(cls, data):
        """Make a list of project object from a list of project's data"""
        return instance_builder(cls, data)

    @classmethod
    def find(cls, pid, api=None):
        """
        Get information about a specific project.

        Args:
            pid (int): project id of the project
            api (RadarlyApi, optional): api which must be used to perform the
                request. If ``None``, the default API will be used.
        Returns:
            Project:
        """
        api = api or RadarlyApi.get_default_api()
        project = api.get(api.router.project['find'].format(project_id=pid))
        return cls(project, api=api)

    def get_distribution(self, parameter, api=None):
        """
        Get distribution of volume, reach, engagement actions
        or/and impressions on a set of publications.

        Args:
            parameter (DistributionParameter): parameter used to specify
                on which subset of publications the distribution must be
                computed and how the distribution must be computed.
                See ``DistributionParameter`` documentation to see
                how to build this object.
            api (RadarlyApi, optional): api which must be used to perform the
                request. If ``None``, the default API will be used.
        Returns:
            Distribution: a list-like object storing all data about the asked
            distribution. You can use the ``pandas`` module to explore this
            object.
        """
        return Distribution.fetch(
            getattr(self, 'id'), parameter, api=api
        )

    def get_publications(self, parameter, api=None):
        """
        Get publications of your project.

        Args:
            parameter (SearchPublicationParameter): parameters object
                made with ``SearchPublicationParameter`` object, which will
                be used as payload data in the POST request to retrieve
                publications.
            api (RadarlyApi, optional): api which must be used to perform the
                request. If ``None``, the default API will be used.
        Returns:
            list[Publication]:
        """
        return Publication.fetch(
            getattr(self, 'id'), parameter, api
        )

    def get_all_publications(self, parameter, api=None):
        """Get all publications matching given parameters. It returns a
        generator which yields publications.

        Args:
            parameter (SearchPublicationParameter): parameters object
                made with the SearchPublicationParameter instance, which will
                be used as payload data in POST request. Go to
                ``SearchPublicationParameter`` documentation to know how to
                build this object.
            api (RadarlyApi, optional): api which must be used to perform the
                request. If ``None``, the default API will be used.
        Returns:
            PublicationGenerator: generator of publications. On each iterations, a
            Publication is yielded until there is no more publication.
        """
        return Publication.fetch_all(
            getattr(self, 'id'), parameter, api
        )

    def get_influencers(self, parameter, api=None):
        """Get influencers linked to a project.

        Args:
            parameter (InfluencerParameter): parameter sent as payload
                to the API. Go to ``InfluencerParameter`` documentation to see
                how to build this object.
            api (RadarlyApi, optional): api which must be used to perform the
                request. If ``None``, the default API will be used.
        Returns:
            list[Influencer]:
        """
        return Influencer.fetch(
            getattr(self, 'id'), parameter, api
        )

    def get_all_influencers(self, parameter, api=None):
        """Get all influencers in a project matching some parameters.

        Args:
            parameter (InfluencerParameter): parameters object which
                is the payload of the request made to get data. Must contains
                pagination's parameter.
            api (RadarlyApi, optional): api which must be used to perform the
                request. If ``None``, the default API will be used.
        Returns:
            InflencersGenerator: generator which yields influencer
        """
        return Influencer.fetch_all(
            getattr(self, 'id'), parameter, api=api
        )

    def get_analytics(self, parameter, api=None):
        """Retrieve some statistics from the API. It allows you to dive deeper
        into the analysis of your publication's project by retrieving several
        kind of analytics, computed on all or a subset of the publications
        stored in your project.

        Args:
            parameter (AnalyticsParameter): parameter used to specify
                which analytics you want to compute, on which subset of
                publications to work... See ``AnalyticsParameter``
                documentation to get some help on how build this object.
            api (RadarlyApi, optional): api which must be used to perform the
                request. If ``None``, the default API will be used.
        Returns:
            Analytics: object storing data retrieved from the API. Each field
            can be explored with ``pandas``.
        """
        focuses = id_to_value(getattr(self, 'focuses'), 'focuses')
        return Analytics.fetch(getattr(self, 'id'), parameter,
                               focuses=focuses, api=api)

    def get_localizations(self, parameter, api=None):
        """Get geographical distribution by region or town.

        Args:
            parameter (LocalizationParameter): object sent as payload to
                the API. Go to ``LocalizationParameter`` documentation to
                know how to build this object.
            api (RadarlyApi, optional): api which must be used to perform the
                request. If ``None``, the default API will be used.
        Returns:
            Localization: list of statistics by geographical point
        """
        return Localization.fetch(getattr(self, 'id'), parameter,
                                  api=api)

    def get_cloud(self, parameter, api=None):
        """Retrieve cloud statistics from the Radarly API.

        Args:
            parameter (CloudParameter): parameter used to
                specify on which subset of publications the cloud
                computations must be performed and which key must
                be returned. Go to ``CloudParameter`` documentation
                to know how you can build this object.
        Returns:
            Cloud: dict-like object storing statistics by fields
        """
        return Cloud.fetch(getattr(self, 'id'), parameter, api)

    def get_pivot_table(self, parameter, api=None):
        """Retrieve data from the API in order to build the pivot table object.

        Args:
            parameter (PivotParameter): parameter used to restrict the set
                of publications on which the pivot table will be based.
                Go to ``PivotParameter`` documentation to see how to build
                this object.
            api (RadarlyApi, optional): api which must be used to perform the
                request. If ``None``, the default API will be used.
        Returns:
            PivotTable: data from the pivot table. You can use ``pandas`` to
            interact with it.
        """
        focuses = id_to_value(getattr(self, 'focuses'), 'focuses')
        fields = id_to_value(getattr(self, 'tags'), 'tags')
        return PivotTable.fetch(getattr(self, 'id'), parameter,
                                focuses=focuses, fields=fields, api=api)

    def get_social_performance(self, parameter, api=None):
        """Retrieve information about social account performance from the API.

        Args:
            parameter (SocialPerformanceParameter): object sent as
                payload to the API. Go to ``SocialPerformanceParameter`` to
                know how to build this object.
            api (RadarlyApi, optional): api which must be used to perform the
                request. If ``None``, the default API will be used.
        Returns:
            SocialPerformance: list-like object compatible with ``pandas``
        """
        return SocialPerformance.fetch(getattr(self, 'id'), parameter,
                                       api=api)

    def get_benchmark(self, parameter, api=None):
        """Retrieve benchmark information from the Radarly API.

        Args:
            parameter (BenchmarkParameter): parameter used to configure
                the benchmark which will be performed. Go to
                ``BenchmarkParameter`` documentation to see how you can
                build this object.
            api (RadarlyApi, optional): api which must be used to perform the
                request. If ``None``, the default API will be used.
        Returns:
            Benchmark: dict-like object storing benchmark data by platform
        """
        return Benchmark.fetch(getattr(self, 'id'), parameter,
                               api=api)

    def get_topic_and_entity(self, parameter, api=None):
        """
        Get distribution by category and subcategory.

        Args:
            parameter (TopicParameter): parameter sent as payload
                to the API. Go to ``TopicParameter`` documenation to know
                how to build this object.
            api (RadarlyApi, optional): api which must be used to perform the
                request. If ``None``, the default API will be used.
        Returns:
            TopicWheel, Entities:
        """
        api = api or RadarlyApi.get_default_api()
        param = parameter.copy()
        url = api.router.topicwheel['fetch'].format(
            project_id=getattr(self, 'id')
        )
        url = url + '?locale={}'.format(param.pop('locale'))
        res_data = api.post(url, data=param)
        categories = []
        entities = []

        for topic in res_data['topics']:
            category = topic['category']
            terms = category.pop('term').split('.')
            category['category'], category['subcategory'] = terms
            category.update(category.pop('counts'))
            categories.append(category)
            for entity in topic['entity']:
                entity.update(entity.pop('counts'))
                entity['category'], entity['subcategory'] = terms
                entities.append(entity)
        categories.sort(key=lambda x: x['category'])
        entities.sort(key=lambda x: x['category'])

        return TopicWheel(categories), Entities(entities)

    def get_geogrid(self, parameter, api=None):
        """retrieve geographical distribution from the API.

        Args:
            parameter (GeoParameter): parameter to specify how to
                compute the geographical distribution. Go to ``GeoParameter``
                documentation to know how to build this object.
            api (RadarlyApi, optional): api which must be used to perform the
                request. If ``None``, the default API will be used.
        Returns:
            GeoGrid: list-like object compatible with ``pandas``
        """
        return GeoGrid.fetch(getattr(self, 'id'), parameter,
                             api=api)

    def get_clusters(self, parameter, api=None):
        """Retrieve clusters from the Radarly API.

        Args:
            parameter (ClusterParameter): parameter to configure
                how the cluster are built and which subet of publications must
                be used. Go to ``ClusterParameter`` documentation to know how
                to build this object.
            api (RadarlyApi, optional): api which must be used to perform the
                request. If ``None``, the default API will be used.
        Returns:
            Cluster: cluster object
        """
        return Cluster.fetch(getattr(self, 'id'), parameter,
                             api=api)


class InfoProject(SourceModel):
    """Object storing information about a project (but not all available
    information). To get the full version of the object, you can use the
    ``expand`` class method.
    """
    def __init__(self, data, translator=None):
        super().__init__(data, translator=translator)

    def __repr__(self):
        pid, label = getattr(self, 'id'), getattr(self, 'label')
        return '<InfoProject.pid={}.label={}>'.format(pid, label)

    @classmethod
    def _builder(cls, data):
        """Make a list of project object from a list of project's data which
        yields project object"""
        return [cls(item) for item in data]

    def expand(self, api=None):
        """Retrieve additional information about a project from the API.

        Args:
            api (RadarlyApi, optional): API used to perform the request.
                If ``None``, the default API will be used.
        Returns:
            Project: object storing all available information about a project
        """
        api = api or RadarlyApi.get_default_api()
        return Project.find(pid=getattr(self, 'id'), api=api)



"""
Useful functions to perform efficient data mining's tasks on a dictionary
"""

import operator
import re

from .colorizer import blue, red
from .misc import extract_type


SEPARATOR_DPATH = re.compile(r'(.(?P<key>[\w]+)'
                             r'(\[(?P<index>[-\d]+)\])?'
                             r'(\((?P<filters>[ #&\(\)*+,.!/:;<=>?@\[\\\]^_`\{|\}~/,\w-]+)\))?)')
CORRECT_DPATH = re.compile(r'(^(\$|\$\$))({})*$'.format(SEPARATOR_DPATH.pattern))


def draw_structure(element, max_depth=2, indent=2, show_type=True, level=0):
    """
    Recursive function to draw the structure of a dictionary.

    Args:
        element (tuple): tuple where the first element is the name of the
            object to examine and the second one is the element itself
        max_depth (int, optional): deeper level to explore. Default
            to 1.
        indent (int, optional): indentation of each level drawn. Default
            to 2
        show_type (bool, optional): whether or not display the type of
            element
        level (int, optional): current level of exploration. Internal
            use.
    Returns:
        None:
    """
    indent_spaces = ("|" + " " * (indent - 1)) * level
    name, data = element
    data_type = extract_type(data)

    if name:
        line = '{}{}'.format(indent_spaces, name)
        if show_type:
            if data_type == 'None':
                data_type = red('({})'.format(data_type))
            elif data_type == 'list' and data:
                first_type = extract_type(data[0])
                data_type = blue('({}[{}])'.format(data_type, first_type))
            else:
                data_type = blue('({})'.format(data_type))
            line += ' {}'.format(data_type)
        print(line)

    if level == max_depth: return None
    is_radarly_object = 'radarly' in str(type(data))
    if isinstance(data, list) and data:
        draw_structure(element=(None, data[0]),
                       level=level,
                       max_depth=max_depth,
                       indent=indent,
                       show_type=show_type)
    elif isinstance(data, dict) or is_radarly_object:
        if is_radarly_object:
            data = dict((key, getattr(data, key)) for key in data.keys())
        for key in sorted(data.keys()):
            draw_structure(element=(key, data[key]),
                           level=level + 1,
                           max_depth=max_depth,
                           indent=indent,
                           show_type=show_type)

    return None


def not_contains(container, element):
    """Antonym of :func:`operator.contains`"""
    return not operator.contains(container, element)


def parse_filter(filters):
    """Parse a filter string, like ``id=323``.

    Args:
        filters (str): string to parse
    Returns:
        dict: dictionary with 3 keys
            *op*
              function used to compare two elements. The operator in the
              filter string is mapped to a function of the `operator` module
            *key*
              key of the object which should be examined (first argument
              of *op* function)
            *value*
              value used to
    """
    if filters is None: return None
    pattern = r' #&\(\)*+,.!/:;<=>?@\[\\\]^_`\{|\}~/,\w-'
    key_pattern = r'^(?P<key>[{}]*)'.format(pattern)
    operator_pattern = r'(?P<operator>=|!=|<|>|>=|<=| in | notin )'
    value_pattern = r'(?P<value>[{}]*)$'.format(pattern)
    filter_pattern = re.compile(key_pattern + operator_pattern + value_pattern)
    match = filter_pattern.match(filters)

    op_dict = {
        '=': operator.eq, '==': operator.eq, '!=': operator.ne,
        '>': operator.gt, '>=': operator.ge,
        '<': operator.lt, '<=': operator.le,
        ' in ': operator.contains, ' notin ': not_contains
    }
    op_func = op_dict[match.group('operator')]

    value = match.group('value')
    if value == 'True': value = True
    if value == 'False': value = False
    try:
        value = float(value)
    except ValueError:
        pass

    if op_func in [operator.contains, not_contains]:
        return dict(value=match.group('key'), key=value.lower(), op=op_func)
    return dict(key=match.group('key'), value=value, op=op_func)


def return_item(item, key, silently=False):
    """Depending of the type of item, return the value corresponding to `key`
    or the attribute `key` of the object item.

    Args:
        item (dict or object): Python object to explore
        key (str): key to search in the object
        silently (bool, optional): whether or not the function should
            raises an error if key is not found. Default to False
    Returns:
        object:
    """
    if isinstance(item, dict):
        return item.get(key, None) if silently else item[key]
    return getattr(item, key, None) if silently else getattr(item, key)


def dpath(source, path):
    """Quickly explore and find specific values in an object by following a
    path. This path must respect some format rules in order to be correctly
    understood by the function.

    * The path must begin with ``$``. If the path starts with ``$$``, the
      function will not raise an error if a key specified in the path is not
      found but will return None instead. If the ``path`` doesn't start with
      ``$``, the behaviour of the function is the same as the standard
      ``__getitem__`` function.
    * Each key of the path must be separated with a point ``.``
    * If you want to get only one element of a list, you can specify the index
      of the element, as in ``dashboards[3]``.
    * You can filter a list to only retrieve items matching a criteria. The
      criteria must be between parenthesis. The supported operator are: =, !=,
      <=, >=, >, <, in, notin

    **Example of valid paths**:

    * ``$.focuses[3].id``
    * ``$$.projects(id>50).limits.stop``
    * ``$.projects(DEMO in label).id``

    **Example of not valid paths**:

    * ``focuses[3].id`` (doesn't start with ``$``)

    Args:
        source (dict): object to explore
        path (str): path conformed to a JSON path
    Returns:
        object: object matching the dpath.
    """
    def get(element, path, silently=False):
        """Core function of the dpath function.

        Ars:
            element (object): element to explore.
            path (list[dict]): path to follow to get the
                element. Each item must be a dictionary with three keys:
                `key` is the name of the object you want to explore (eg
                a key of a dictionary or an attribute of an object),
                `filters` is a dictionary specifying how to filter a list
                (can be None) and `index` is the index of an element in a
                list (can be None)
            silently (bool, optional): whether or not the function
                should raises an error if the key in not found. Default
                to False.
        Returns:
            object:
        """
        def get_item(item, key, index, filters):
            """Get a value matching a key in an object."""
            right_item = return_item(item, key, silently)
            if isinstance(right_item, list) and index and right_item:
                return right_item[int(index)]
            if isinstance(right_item, list) and filters and right_item:
                right_item = [
                    item for item in right_item
                    if filters['op'](return_item(item, filters['key'], silently),
                                     filters['value'])
                ]
                if len(right_item) == 1: return right_item[0]
                return right_item
            return right_item

        if path and element:
            if isinstance(element, list):
                return [
                    get(get_item(item, **path[0]), path[1:], silently) for item in element
                ]
            return get(get_item(element, **path[0]), path[1:], silently)
        return element

    if CORRECT_DPATH.match(path):
        silently = path.startswith('$$')
        path = [
            dict(
                key=item.group('key'),
                index=item.group('index'),
                filters=parse_filter(item.group('filters')),
            ) for item in SEPARATOR_DPATH.finditer(path)
        ]
        return get(source, path, silently=silently)
    path = dict(
        key=path,
        index=None,
        filters=None,
    )
    return get(source, [path])



"""
This module defines objects used to explore information about a Radarly User.
For security reasons, you can only retrieve information about the
current user of the API.
"""

from .api import RadarlyApi
from .model import SourceModel
from .project import InfoProject


class User(SourceModel):
    """Object used to explore user information returned by the API. Given that
    this object inherits from the ``SourceModel``, you can get the structure of the
    object with the ``draw_structure`` method.

    Examples:
        >>> user = User.find(uid='me')
        >>> user
        <User.id=1234.email='john.doe@linkfluence.com'>
        >>> user.keys()
        {'projects', 'current_project_id', ..., 'is_disabled'}

    Args:
        id (int): unique identifier of the user
        name (str): registred name of the user
        email (str): regitred email of the user
        projects (list[InfoProject]): list in which each item
            is an object storing some information about a project (notice
            that all information about a project are not stored in this
            object)
        created (datetime.datetime): creation datetime of the user
        ...
    """
    def __init__(self, data):
        super().__init__()
        translator = dict(
            projects=InfoProject._builder,
        )
        super().add_data(data, translator)

    def __repr__(self):
        uid, email = self['id'], self['email']
        return "<User.id={}.email='{}'>".format(uid, email)

    @classmethod
    def find(cls, uid, api=None):
        """
        Get information about an user.

        Args:
            uid (string): because you can only access data about you, this
                argument must be set to ``me``
            api (RadarlyApi, optional): API used to make the
                request. If None, the default API will be used.
        Returns:
            User: User object storing information retrieved from the API
        """
        api = api or RadarlyApi.get_default_api()
        if uid == 'me':
            user_data = api.get(api.router.user['me'])
            return cls(user_data)
        raise ValueError("The 'uid' argument must be set to 'me'.")



"""
Radarly can compute some statistics and then group the result by countries or
town. The ``localization`` modules defines all objects in order to interact
with such results.
"""

import copy

from .api import RadarlyApi
from .utils.misc import to_snake_case


class Localization(list):
    """List-like object used to store data about the geographical
    distribution. Each item of the distribution is a dictionary of statistics
    about a specific localization. This object is compatible with ``pandas``,
    so you convert it to a DataFrame:

    >>> import pandas as pd
    >>> localizations
    Localization(length=100)
    >>> pd.DataFrame(localizations).head()
          doc                 id              idStr isoCode lang       lat  \\
    0   20963  72057594041390308  72057594041390308   BR-GO   en -15.58071
    1   38483  72057594043029696  72057594043029696   US-NJ   en  40.16706
    2  551932  72057594044197067  72057594044197067  GB-ENG   en  52.16045
    3   16393  72057594042289821  72057594042289821   US-MD   en  39.00039
    4   35604  72057594040566296  72057594040566296  GB-SCT   en  56.00000
    ...
    """
    def __init__(self, data):
        super().__init__()
        for item in (i for i in data):
            element = item['counts']
            element.update(item['info'])
            del item['counts']
            del item['info']
            element.update(item)
            self.append(element)

    def __repr__(self):
        return '<Localization.length={}>'.format(len(self))

    @classmethod
    def fetch(cls, project_id, parameter, api=None):
        """Get geographical distribution by country or town.

        Args:
            project_id (int): identifier of your project
            parameter (LocalizationParameter): object sent as payload to
                the API. See ``LocalizationParameter`` on how to build this
                object.
            api (RadarlyApi, optional): API used to make the
                request. If None, the default API will be used.
        Returns:
            Localization: list of stats by geographical point
        """
        parameter = copy.deepcopy(parameter)
        api = api or RadarlyApi.get_default_api()
        url = api.router.localization['fetch'].format(
            project_id=project_id,
            region_type=parameter.pop('geo_type', 'region')
        )
        params = [
            ('locale', parameter.pop('locale', 'en_GB'))
        ]
        data = api.post(url, params=params, data=parameter)
        return cls(data[to_snake_case('geo-digging')])



"""
The topic wheel is a distribution of publications grouped by categories.
"""


class TopicWheel(list):
    """List-like object storing information about the topics."""
    def __init__(self, data):
        super().__init__()
        self.extend(data)

    def __repr__(self):
        return '<Topics.length={}>'.format(len(self))


class Entities(list):
    """List-like object storing information about entities"""
    def __init__(self, data):
        super().__init__()
        self.extend(data)

    def __repr__(self):
        return '<Entities.length={}>'.format(len(self))



"""
In Radarly, the cloud features is used to visualize the most relevant keywords,
hashtags, mentions, named entities... associated to your project or set of
publications.

The ``Cloud`` object in ``radarly`` is built to give you the possibility to retrieve
this kind of data.
"""

from .api import RadarlyApi


class Cloud(dict):
    """Dict-like object used to store the data of the cloud. The keys of this
    object corresponds to the differents statistics types available for the
    cloud (e.g *keywords*, *hashtags*, *mentions*,...). The value associated
    to each key of the ``Cloud`` object can easily be converted into a
    ``DataFrame``.

    Example:

        >>> import pandas as pd
        >>> cloud
        <Cloud.fields=['emojiCharts', 'mentions',... , 'hashtags', 'affects']>
        >>> mentions = pd.DataFrame(cloud['mentions'])
        >>> mentions
                            doc  impression     reach
        ABHcosmetics   10637    23040848   1843259
        BabyAnnie218       0      204310      8170
        BenefitBeauty   1997    16217028   1297354
        CHANEL          6207   706216524  56497312
        CHANELBEAUTY      34    48025113   1920999
    """
    def __init__(self, data):
        super().__init__()
        self.update(data)

    def __repr__(self):
        return '<Cloud.fields={}>'.format(list(self.keys()))

    @classmethod
    def fetch(cls, project_id, parameter, api=None):
        """Retrieve cloud data from the Radarly API.

        Args:
            project_id (int): identifier of your project.
            parameter (CloudParameter): parameter used to
                specify on which subset of publications the cloud
                computations must be performed and which key must
                be returned. See ``CloudParameter`` documentation
                to know how you can build this object.
            api (RadarlyApi, optional): API used to perform the
                request. If None, the default API will be used.
        Returns:
            Cloud: dict-like object storing statistics by fields
        """
        api = api or RadarlyApi.get_default_api()
        url = api.router.cloud['fetch'].format(project_id=project_id)
        data = api.post(url, data=parameter)
        return cls(data['cloud'])



"""
In Radarly, the benchmark feature allows you to compare your own social media
accounts with those of your competitors. Thanks to that, you can manage and
measure the performance of your accounts.
This module defines the Python object used to store the data of this
benchmark.
"""

from .api import RadarlyApi


class Benchmark(dict):
    """Dict-like object used to store the benchmark data. The keys of this
    object are the available platforms on which the different social accounts
    have been compared. The value associated to each platform is an object
    which can be transformed into a ``pandas.DataFrame`` for a better
    exploration.

    Example:

        >>> benchmark
        Benchmark(platforms=['facebook', 'instagram'])
        >>> import pandas as pd
        >>> pd.DataFrame(benchmark['facebook'])
    """

    def __init__(self, data):
        super().__init__()
        for platform in data:
            self[platform] = []
            for account in data[platform]:
                for stat in account['stats']:
                    item = {'date': stat['date']}
                    item.update(stat['scores'])
                    item.update({'account_id': account['uid']})
                    item['platform'] = platform
                    self[platform].append(item)

    def __repr__(self):
        return '<Benchmark.platforms={}>'.format(list(self.keys()))

    @classmethod
    def fetch(cls, project_id, parameter, api=None):
        """Retrieve benchmark information from the Radarly API.

        Args:
            project_id (int): identifier of your project where information is
                stored
            parameter (BenchmarkParameter): parameter used to configure
                the benchmark which will be performed. See the documentation
                of ``BenchmarkParameter`` to see how you can build this object.
            api (RadarlyApi, optional): API object used to perform the request.
                If None, it will use the default API.
        Returns:
            Benchmark: dict-like object storing benchmark data by platform
        """
        api = api or RadarlyApi.get_default_api()
        url = api.router.benchmark['fetch'].format(project_id=project_id)
        data = api.get(url, params=parameter)
        return cls(data)



"""A cluster in Radarly is a set of publications grouped by strong
similarities (for example, articles published in several medias but
with the same source will be grouped in the same cluster)."""

from copy import deepcopy
from reprlib import repr as truncate_repr

from .analytics import Analytics
from .api import RadarlyApi
from .model import SourceModel
from .publication import Publication
from .cloud import Cloud


class Cluster(SourceModel):
    """Object which inherits from the ``SourceModel`` storing information
    about a cluster"""
    def __init__(self, data, api=None):
        self._api = api or RadarlyApi.get_default_api()
        del data['stats']
        super().__init__(data)

    def __repr__(self):
        return '<Cluster.size={}.story={}>'.format(
            getattr(self, 'size'), truncate_repr(getattr(self, 'story'))
        )

    @classmethod
    def fetch(cls, project_id, parameter, api=None):
        """Retrieve clusters from the Radarly API.

        Args:
            project_id (int): identifier of your project
            parameter (ClusterParameter): parameter to configure
                how the cluster are calculated. See ``ClusterParameter``
                documentation to know how build this object.
            api (RadarlyApi, optional): API used to perform the
                request. If None, the default API will be used.
        Returns:
            Cluster: cluster object (you can the ``draw_structure`` to
                get its structure)
        """
        api = api or RadarlyApi.get_default_api()
        url = api.router.cluster['fetch'].format(project_id=project_id)
        data = api.post(url, data=parameter)
        ans = []
        for data in data['hits']:
            data['project_id'] = project_id
            ans.append(cls(data, api=api))
        return ans

    def get_analytics(self, parameter, focuses=None):
        """Compute some insights about the set of publications in the
        cluster.

        Args:
            parameter (AnalyticsParameter): object sent as payload to the
                API to configure insights computation. See
                ``AnalyticsParameter`` documentation to see how build this
                object.
            focuses (dict, optional): dictionary used to translate focuses
                if 'focuses' was asked as field
        Returns:
            Analytics:
        """
        parameter = deepcopy(parameter)
        parameter['stories'] = [getattr(self, 'story')]
        return Analytics.fetch(getattr(self, 'project_id'), parameter,
                               focuses=focuses, api=self._api)

    def get_publications(self, parameter):
        """Retrieve publications in the cluster.

        Args:
            parameter (SearchPublicationParameter): object sent as payload to
                the API to configure insights computation. See
                ``SearchPublicationParameter`` documentation to see how to
                build this object.
        Returns:
            Analytics:
        """
        parameter = deepcopy(parameter)
        parameter['stories'] = [getattr(self, 'story')]
        return Publication.fetch(
            getattr(self, 'project_id'), parameter, self._api
        )

    def get_cloud(self, parameter):
        """Retrieve clouds insights about publications in the cluster.

        Args:
            parameter (CloudParameter): object sent as payload to the API. See
                ``CloudParameter`` documentation to see how to build this
                object.
        Returns:
            Cloud:
        """
        parameter = deepcopy(parameter)
        parameter['stories'] = [getattr(self, 'story')]
        return Cloud.fetch(
            getattr(self, 'project_id'), parameter, self._api
        )



"""
The pivot table is a Radarly's feature which allows you to perform
cross data analysis.
"""

from .api import RadarlyApi
from .utils._internal import CallableDict, id_to_value
from .utils.misc import to_snake_case


class PivotTable(dict):
    """Dict-like object storing data of the pivot table.
    It is strongly advised to use ``pandas`` to explore this object.

    >>> import pandas as pd
    >>> table
    PivotTable(pivot='Elegance', against='focuses')
    >>> df_table = pd.DataFrame(table)
    >>> df_table.head()
                             CHANEL F&B  Dior F&B
    doc ARMANI Fragrances          6722      5196
        BURBERRY Fragrances        5472      3494
        CHANEL Fragrances        130240     21481
        CLINIQUE Fragrances        1150       835
        DIOR Fragrances           21516     83880
    """
    def __init__(self, data, focuses=None, fields=None):
        super().__init__()
        if focuses is None: focuses = {}
        if fields is None: fields = {}
        translator = CallableDict(focuses, fields)
        self.pivot = data['pivot']
        self.against = data['against']
        for focus in data[to_snake_case(self.pivot)]:
            for element in focus[to_snake_case(self.against)]:
                focus_id = element['term'].replace('focus_', '')
                term = translator(focus_id)
                self.setdefault(term, dict())
                self[term].update({
                    (metric, translator.get(
                        focus['term'].replace('focus_', ''), focus['term'])
                    ): element['counts'][metric]
                    for metric in focus['counts']
                })

    def __repr__(self):
        return '<PivotTable.pivot={0.pivot}.against={0.against}>'.format(self)

    @classmethod
    def fetch(cls, project_id, parameter,
              autotranslate=False, focuses=None, fields=None,
              api=None):
        """Retrieve data from the API in order to build the pivot table object.

        Args:
            project_id (int): identifier of the project
            parameter (PivotParameter): parameter used to restrict data
                used to build the pivot table. See ``PivotParameter``
                documentation to see how to build this object.
            autotranslate (bool): whether or not translate the headers and the
                index in order to have labels and not integer ids. If True, it
                will consume one extra request.
            focuses (dict, optional): dictionary used to translate the headers
                (if *against* is set to 'focus' in the parameter)
            fields (dict, optional): dictionary used to translate the index
            api (RadarlyApi, optional): API used to make the
                request. If None, the default API will be used.
        Returns:
            PivotTable: data from the pivot table. You can use ``pandas`` to
            interact with it.
        """
        api = api or RadarlyApi.get_default_api()
        url = api.router.pivot_table['fetch'].format(project_id=project_id)
        data = api.post(url, data=parameter)

        if autotranslate and not focuses and not fields:
            from .project import Project
            project = Project.find(pid=project_id)
            focuses = id_to_value(getattr(project, 'focuses'), 'focuses')
            fields = id_to_value(getattr(project, 'tags'), 'tags')

        return cls(data, focuses, fields)



"""
Radarly gives you an option in order to get the geographical distribution
of your publications. You can also compute some advanced search using, for
example, a polygon made of coordinates in order to get statistics in a specific
region.
"""

from .api import RadarlyApi


class GeoGrid(list):
    """List-like object used to store the data related to the geographical
    disitribution. Each item corresponds to a geographical point. This object
    is compatible with ``pandas``.

    Example:

        >>> import pandas as pd
        >>> geo_distribution
        GeoGrid(nb_points=1004)
        >>> df_geo = pd.DataFrame(geo_distribution)
        >>> df_geo.head()
            doc geohash       lat        lon
        0  2857     u09  48.85482    2.33683
        1  1995     ucf  55.74689   37.58481
        2   976     dr5  40.68962  -74.01365
        3   969     gcp  51.47043   -0.20134
        4   851     wec  22.30178  114.15957
    """
    def __init__(self, data):
        super().__init__()
        for geoitem in data:
            geoitem.update(geoitem.pop('centroid'))
            geoitem.update(geoitem.pop('counts'))
            self.append(geoitem)

    def __repr__(self):
        return '<GeoGrid.length={}>'.format(len(self))

    @classmethod
    def fetch(cls, project_id, parameter, api=None):
        """retrieve geographical distribution from the API.

        Args:
            project_id (int): unique identifier of the project
            parameter (GeoParameter): parameter to specify how to
                compute the geographical distribution and on which subset of
                publications work. See ``GeoParameter`` documentation to see
                how to build this object.
            api (RadarlyApi, optional): API used to make the
                request. If None, the default API will be used.
        Returns:
            GeoGrid: list-like object compatible with ``pandas``
        """
        api = api or RadarlyApi.get_default_api()
        url = api.router.geogrid['fetch'].format(project_id=project_id)
        data = api.post(url, data=parameter)
        return cls(data)



"""
Miscellaneous functions and classes
"""

import json
import re
import textwrap
from collections import UserList, namedtuple
from datetime import datetime
from functools import reduce
from os.path import dirname, join, pardir
from urllib.parse import urlparse

from dateutil.relativedelta import relativedelta
from dateutil.rrule import MONTHLY, rrule


def load_data(filepath):
    """Load static file located in the package directory"""
    filepath = join(dirname(__file__), pardir, filepath)
    with open(filepath, mode='r') as data_file:
        data = json.load(data_file)
    return data


def dict_to_namedtuple(name, data):
    """Converts a dictionary into a namedtuple"""
    return namedtuple(name, data.keys())(**data)


flat = lambda x: reduce(lambda a, b: a+b, x, [])


class Quarter(UserList):
    """List object storing start date and end date of a quarter"""
    def __init__(self, n_quarter, year):
        super().__init__()
        self.n_quarter = n_quarter
        self.year = year
        start = datetime(year, 1, 1)
        start_quarter = list(
            rrule(MONTHLY, interval=3, dtstart=start, count=4)
        )[n_quarter - 1]
        end_quarter = start_quarter + relativedelta(months=3, days=-1)
        self.data.extend([start_quarter, end_quarter])

    def __contains__(self, value):
        return value >= self[0] and value <= self[1]

    @classmethod
    def of_year(cls, year):
        """Generates all Quarter object of a specific year"""
        start = datetime(year, 1, 1)
        start_quarter = list(
            rrule(MONTHLY, interval=3, dtstart=start, count=4)
        )
        end_quarter = [
            date + relativedelta(months=3, days=-1) for date in start_quarter
        ]
        return [cls(*item) for item in list(zip(start_quarter, end_quarter))]

    def __repr__(self):
        return '<Quarter.nth={}.year={}>'.format(
            1 + self.data[0].month // 3,
            self.data[0].year
        )


def parse_date(date_string):
    """Parse a date string with the format '%Y-%m-%dT%H:%M:%S.%fZ'.

    Args:
        date_string (str): string with the format '%Y-%m-%dT%H:%M:%S.%fZ'
    Returns:
        datetime:
    """
    return datetime.strptime(date_string, '%Y-%m-%dT%H:%M:%S.%fZ')


def to_snake_case(name):
    """Converts a string into snake_case format"""
    name = name.replace('-', '_')
    temp = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', temp).lower()


def extract_type(obj):
    """Get the name of a Python object, based on its type

    Args:
        obj (object): Python object
    Returns:
        str:
    """
    pattern = re.compile(r"^<class '(?P<obj_type>[.a-zA-Z/_]*)'>$")
    match = pattern.match(str(type(obj)))
    ans = match.group('obj_type') if match else str(type(obj))
    translator = dict(NoneType='None')
    ans = ans.split(".")[-1]
    return translator.get(ans, ans)


def dumps_datetime(date):
    """Datetime into string"""
    if isinstance(date, datetime):
        return date.__str__()
    return None


def parse_image_url(image_url):
    """Parse the image filename using a regular expression.

    Args:
        image_url (string): url which must be parsed
    Returns:
        tuple (str or None): filename and format got from the filename.
        If it doesn't match, returns a None tuple.
    """
    path = urlparse(image_url).path
    pattern = re.compile(
        r'/(?P<filename>[0-9a-zA-Z_-]*).(?P<format>[a-z0-9]{3,4})$'
    )
    match = pattern.search(path)
    if match:
        return match.group('filename'), match.group('format')
    return None, None


def dict_to_tsv(data):
    """Get a pretty representation of a dictionary as a table.

    Args:
        data_error (dict):
    Returns:
        str: pretty representation of the dictionary ``data``
    """
    offset = max([len(item) for item in data.keys()]) + 3

    def multiline(long_string):
        """Wraps a string and add some indentation to lines (except
        for the first one).

        Args:
            long_string (str): string to wrap
        Returns:
            str: string wrapped.
        """
        multilines = textwrap.wrap(long_string, width=70)
        multilines = [
            int(index > 0) * offset * ' ' + item
            for index, item in enumerate(multilines)
        ]
        return "\n".join(multilines)

    template = []
    for item in data:
        multilined = multiline(str(data[item]))
        template.append("{label:{offset}}{text}".format(
            label=item, text=multilined, offset=offset))
    return "\n".join(template)



"""
Used to control Python object type
"""

import pycountry
from dateutil.parser import parse

from .misc import flat


def check_date(date_object):
    """Convert a date into the right for API"""
    if date_object is None: return None
    if isinstance(date_object, str):
        date_object = parse(date_object)
    date_object = date_object.isoformat()
    return date_object


def check_language(language):
    """Check if a language code or object is correct"""
    if isinstance(language, (list, tuple)):
        return flat([check_language(item) for item in language])
    if language == 'xx': return ['xx']
    if language == 'zh': return ['zh-cn', 'zh-tw']
    if language in ['zh-cn', 'zh-tw']:
        return [language]
    try:
        ans = pycountry.languages.lookup(language).alpha_2
    except LookupError:
        raise ValueError("{} is an unknown language code.".format(
            language
        ))
    return [ans]


def check_geocode(country):
    """Check if a country code or object is correct"""
    if isinstance(country, (list, tuple)):
        return [check_geocode(item) for item in country]
    if country == 'xx': return 'xx'
    try:
        ans = pycountry.countries.lookup(country).alpha_2
    except LookupError:
        raise ValueError("{} is an unknown geographical code.".format(
            country
        ))
    return ans


def check_list(elements, rtype, assertion_message=None):
    """Check if all the items of a list has the right type"""
    error_message = "Invalid type (must be a list of {})".format(rtype)
    assertion_message = assertion_message or error_message
    assert isinstance(elements, list), 'A list must be given.'
    assert all([isinstance(element, rtype) for element in elements]), \
        assertion_message
    return None



"""
The aim of this module is to give you the possibility to make some
computations on your publications. Each analytics is returned as a time
distribution of some metrics.

.. note:: Each time you try to retrieve some analytics, the Radarly's server
    launch some computation tasks, hence the loading time which can reach
    several seconds.
"""

from .api import RadarlyApi
from .constants import ANALYTICS_FIELD
from .utils._internal import CallableDict


class Analytics(dict):
    """Dict-like object. Each key gives access to a specific kind of
    statistics (asked in the parameter in the *fields* key).
    This object can be explored with ``pandas``.

    >>> import pandas as pd
    >>> stats
    Analytics(fields=['occupations', 'languages', ..., 'logos', 'genders'])
    >>> languages = pd.DataFrame(stats['languages'])
    >>> languages.head()
                    af   ar  ar-bz  bg  bn  ca  cs   da  de  dv  ...     th  \\
    doc 2018-01-01 NaN   79    1.0   9 NaN  11   6  NaN  22 NaN  ...    214
        2018-01-02 NaN  199    NaN  30 NaN   6   6  7.0  27 NaN  ...    275
        2018-01-03 NaN  132    1.0  13 NaN   4   3  1.0  40 NaN  ...    207
        2018-01-04 NaN  109    NaN  10 NaN   5   3  7.0  43 NaN  ...    212
        2018-01-05 NaN   83    NaN  13 NaN   9   6  8.0  36 NaN  ...    206
    ...

    Args:
        fields: fields currently stored in the Analytics objects. Each value
            associated to a field can be transformed to a DataFrame.
    """
    def __init__(self, data, focuses=None):
        super().__init__()
        focuses = (lambda x: x) if focuses is None else CallableDict(focuses)
        translator = dict(
            focuses=focuses
        )
        self.setdefault('total', {'total': {}})
        self.setdefault('counts', {})
        for dot in data['dots']:
            dot_date = dot['date']

            self['total']['total'][dot_date] = dot['total']
            for major in dot['counts'].keys():
                self['counts'].setdefault(major, {})
                self['counts'][major][dot_date] = dot['counts'][major]

            _ = [self.setdefault(key, {}) for key in dot['stats'].keys()]

            for major in dot['stats'].keys():
                temp = {
                    translator.get(major, lambda x: x)(item['term']): {
                        (metric, dot_date):
                            item['counts'][metric]
                        for metric in item['counts']
                    } for item in dot['stats'][major]
                }
                for field in temp:
                    if not str(field).startswith('tag_offset_'):
                        self[major].setdefault(field, {})
                        self[major][field].update(temp[field])

        self.fields = list(self.keys())

    def __repr__(self):
        return '<Analytics.fields={}>'.format(list(self.keys()))

    @classmethod
    def fetch(cls, project_id, parameter,
              focuses=None, api=None):
        """Retrieve some insights from the API. It allows you to dive deeper
        into the analysis of your project by retrieving several kind of
        analytics, computed on all or a subset of the publications stored in
        your project.

        Args:
            project_id (int): id of your project where all data are stored
            parameter (AnalyticsParameter): parameter used to specify
                which analytics you want to compute, on which subset of
                publications to work... See ``AnalyticsParameter``
                documentation to get some help on how build this object.
            focuses (dict): used to translate headers if *focuses* was asked
                in the field.
            api (RadarlyApi, optional): API to use to made the request. If
                None, it will use the default API.
        Returns:
            Analytics: object storing data retrieved from the API and which
            can be explored with ``pandas``
        """
        api = api or RadarlyApi.get_default_api()
        url = api.router.analytics['global'].format(project_id=project_id)

        is_occupation_asked = ANALYTICS_FIELD.OCCUPATIONS in \
            parameter.get('fields', [])
        if is_occupation_asked:
            parameter['fields'].remove(ANALYTICS_FIELD.OCCUPATIONS)
        if parameter.get('fields', []) == [] and is_occupation_asked:
            data = dict(dots=[])
        else:
            data = api.post(url, data=parameter)
        if is_occupation_asked:
            _ = parameter.pop('fields')
            url = api.router.analytics['occupation'].format(
                project_id=project_id
            )
            occupations_data = api.post(url, data=parameter)
            data['dots'] += occupations_data['dots']

        return cls(data, focuses)



"""
Parameters object sent as payload data.
"""

import json
from urllib.parse import urlencode

from .field import (ClusterMixin, FctxMixin, FieldsMixin, GeoFilterMixin,
                    GeoTypeMixin, IntervalMixin, LocaleMixin, MetricsMixin,
                    PaginationMixin, RangeDateMixin, SortMixin,
                    StandardParameterMixin, TimezoneMixin)

__all__ = (
    'AnalyticsParameter',
    'BenchmarkParameter',
    'CloudParameter',
    'ClusterParameter',
    'DistributionParameter',
    'GeoParameter',
    'InfluencerParameter',
    'LocalizationParameter',
    'PivotParameter',
    'SearchPublicationParameter',
    'SocialPerformanceParameter',
    'TopicParameter',
)


class Parameter(dict):
    """An object based on ``dict`` used to store the payload data sent with
    each request in RadarlyApi. Each road in the API used an object (which
    inherits from this class) to help you build the payload data which will
    be sent to the API. When the methods of Parameter object is used to build
    the object, some checks are computed on the values in order to assert that
    the payload data is correctly built.
    """
    pass


class AnalyticsParameter(Parameter,
                         StandardParameterMixin,
                         FctxMixin,
                         MetricsMixin,
                         FieldsMixin,
                         ClusterMixin):
    """Helper to build the payload sent when retrieving some statistics.
    This helper contains all the methods to restrict the set of publications
    (methods defined in radarly.parameters.field.StandardParameterMixin).

    Example:

        >>> from radarly.constants import INTERVAL, METRIC, PLATFORM
        >>> from radarly.constants import ANALYTICS_FIELD
        >>> start = datetime(2018, 1, 1)
        >>> end = datetime(2018, 1, 31)
        >>> param = AnalyticsParameter() \\
                .metrics(METRIC.DOC, METRIC.IMPRESSION, METRIC.ENGAGEMENT) \\
                .fields(ANALYTICS_FIELD.TONE, ANALYTICS.FOCUSES) \\
                .focuses(include=[1, 2, 3], exclude=[4, 5]) \\
                .fctx(1, 2, 3) \\
                .publication_date(start, end) \\
                .interval(INTERVAL.MONTH) \\
                .followers(1000, 5000) \\
                .media(MEDIA.IMAGE) \\
                .platforms(PLATFORM.INSTAGRAM) \\
                .query('danone')

    """
    pass


class BenchmarkParameter(Parameter,
                         RangeDateMixin,
                         TimezoneMixin,
                         IntervalMixin):
    """Helper to build the payload sent to the API when retrieving benchmark
    data.

    Example:

        >>> start = datetime(2018, 1, 1)
        >>> end = datetime(2018, 1, 31)
        >>> BenchmarkParameter() \\
                .timezone('Europe/Paris') \\
                .entities(1) \\
                .date_range(start, end)
    """
    def entities(self, *entities_ids):
        entities_ids = [str(entities_id) for entities_id in entities_ids]
        self['entities'] = ','.join(entities_ids)
        return self


class CloudParameter(Parameter,
                     StandardParameterMixin,
                     FctxMixin,
                     MetricsMixin,
                     TimezoneMixin,
                     FieldsMixin,
                     ClusterMixin):
    """Helper to build the payload sent to the API when retrieving cloud data.
    This helper contains all the methods to restrict the set of publications
    (methods defined in radarly.parameters.field.StandardParameterMixin).

    Example:

        >>> from radarly.constants import METRIC, PLATFORM
        >>> from radarly.constants import CLOUD_FIELD
        >>> start = datetime(2018, 1, 1)
        >>> end = datetime(2018, 1, 31)
        >>> cloud_param = CloudParameter() \\
                .focuses(include=[1, 2, 3, 4]) \\
                .fctx(*[1, 2, 3, 4]) \\
                .tones(TONE.POSITIVE) \\
                .timezone('Europe/Paris') \\
                .metrics(METRIC.DOC, METRIC.IMPRESSION, METRIC.REACH) \\
                .fields(CLOUD_FIELD.KEYWORDS, CLOUD_FIELD.MENTIONS) \\
                .publication_date(start, end)
    """
    pass


class ClusterParameter(Parameter,
                       StandardParameterMixin,
                       MetricsMixin,
                       SortMixin,
                       PaginationMixin):
    """Helper to build the payload sent to the API when retrieving clusters
    of publications. This helper contains all the methods to restrict the
    set of publications (methods defined in
    radarly.parameters.field.StandardParameterMixin).

    ..warning:: The value for the sort_by parameter are restricted.
        Check the ``AVAILABLE_SORT_BY`` class attribute to know the
        available parameters.
    """
    AVAILABLE_SORT_BY = [
        "volumetry",
        "radar.impression",
        "radar.reach"
    ]

    def __init__(self):
        super().__init__()
        self.flag(retweet=True)


class DistributionParameter(Parameter,
                            StandardParameterMixin,
                            IntervalMixin,
                            MetricsMixin,
                            GeoFilterMixin):
    """Helper to build the payload sent to the API when retrieving time
    distribution of some metrics. This helper contains all the methods
    to restrict the set of publications (methods defined in
    radarly.parameters.field.StandardParameterMixin).

    Example:

        >>> from radarly.constants import METRIC
        >>> start = datetime(2018, 1, 1)
        >>> end = datetime(2018, 1, 31)
        >>> param = DistributionParameter() \\
                .publication_date(start, end) \\
                .focuses(include=[1, 2, 3, 4], exclude=[5]) \\
                .flag(retweet=False) \\
                .metrics(METRIC.DOC, METRIC.IMPRESSION)
    """
    pass


class GeoParameter(Parameter,
                   StandardParameterMixin,
                   MetricsMixin,
                   FctxMixin,
                   GeoFilterMixin,
                   TimezoneMixin):
    """Helper to build the payload sent to the georgrid of the API.
    This helper contains all the methods to restrict the set of publications
    (methods defined in radarly.parameters.field.StandardParameterMixin).

    Example:

        >>> from radarly.constants import METRIC
        >>> start = datetime(2018, 1, 1)
        >>> end = datetime(2018, 1, 31)
        >>> geo_param = GeoParameter() \\
                .focuses(include=[1, 2, 3, 4]) \\
                .fctx(*[1, 2, 3, 4]) \\
                .flag(retweet=False) \\
                .geofilter() \\
                .metrics(METRIC.DOC) \\
                .publication_date(start, end)
    """
    pass


class InfluencerParameter(Parameter,
                          StandardParameterMixin,
                          SortMixin,
                          PaginationMixin):
    """Helper to build the payload when retrieving influencers' data.
    This helper contains all the methods to restrict the set of publications
    (methods defined in radarly.parameters.field.StandardParameterMixin).

    ..warning:: The value for the sort_by parameter are restricted.
        Check the ``AVAILABLE_SORT_BY`` class attribute to know the
        available parameters.

    Example:

        >>> from radarly.constants import PLATFORM
        >>> param = InfluencerParameter() \\
                .platforms(PLATFORM.FACEBOOK) \\
                .focuses(include=[1, 2, 3, 4]) \\
                .publication_date(start, end) \\
                .sort_by('reach') \\
                .pagination(0, 25)
    """
    AVAILABLE_SORT_BY = [
        # 'name',
        'impressions',
        'reach',
        'post'
    ]

    @classmethod
    def default(cls):
        param = cls()
        param = param.sort_by('name').sort_order('desc').pagination(0, 25)
        return param


class LocalizationParameter(Parameter,
                            StandardParameterMixin,
                            LocaleMixin,
                            GeoTypeMixin,
                            TimezoneMixin,
                            MetricsMixin):
    """Helper to build the payload sent to the API when retrieving distribution
    of publications by geographical zones. This helper contains all the methods
    to restrict the set of publications (methods defined in
    radarly.parameters.field.StandardParameterMixin).

    Example:

        >>> from radarly.constants import METRIC
        >>> param = LocalizationParameter() \\
                .locale('en_GB') \\
                .timezone('Europe/Paris') \\
                .metrics(METRIC.DOC)
    """
    pass


class PivotParameter(Parameter,
                     StandardParameterMixin,
                     TimezoneMixin,
                     MetricsMixin,
                     FctxMixin,
                     IntervalMixin):
    """Helper to build the payload for the pivot table road in the API.
    This helper contains all the methods to restrict the set of publications
    (methods defined in radarly.parameters.field.StandardParameterMixin).

    Example:

        >>> from radarly.constants import METRIC, INTERVAL
        >>> start = datetime(2018, 1, 1)
        >>> end = datetime(2018, 1, 31)
        >>> pivot_param = PivotParameter(pivot='Fragrances', against='focuses') \\
                .metrics(METRIC.DOC) \\
                .publication_date(start, end) \\
                .focuses(include=[1, 2, 3, 4]) \\
                .flag(retweet=False) \\
                .fctx(*[1, 2, 3, 4])

    """
    def __init__(self, pivot=None, against=None):
        super().__init__()
        if pivot: self['pivot'] = pivot
        if against: self['against'] = against

    def conf(self, pivot, against):
        self['pivot'] = pivot
        self['against'] = against
        return self


class SearchPublicationParameter(Parameter,
                                 StandardParameterMixin,
                                 SortMixin,
                                 PaginationMixin,
                                 FctxMixin,
                                 ClusterMixin,
                                 GeoFilterMixin):
    """Helper to build the payload sent to the API when you are searching
    pubications. This helper contains all the methods to restrict the set
    of publications (methods defined in
    radarly.parameters.field.StandardParameterMixin).

    ..warning:: The value for the sort_by parameter are restricted.
        Check the ``AVAILABLE_SORT_BY`` class attribute to know the
        available parameters.

    Example:

        >>> start = datetime(2018, 1, 1)
        >>> end = datetime(2018, 1, 31)
        >>> param = SearchPublicationParameter() \\
                .publication_date(start, end) \\
                .platforms(PLATFORM.INSTAGRAM) \\
                .languages('french', 'english', 'zh-cn', 'zh-tw') \\
                .flag(retweet=False, favorite=True) \\
                .tones(TONE.POSITIVE, TONE.NEUTRAL) \\
                .sort_by(BY.ENGAGEMENT) \\
                .pagination(start=0, limit=25)
    """
    AVAILABLE_SORT_BY = [
        'date',
        'radar.virality',
        'radar.engagement',
        'radar.reach',
        'radar.impression',
        'radar.rating',
        'random'
    ]

    @classmethod
    def default(cls):
        param = cls()
        param = param.sort_by('date').sort_order('desc').pagination(0, 25)
        return param


class SocialPerformanceParameter(Parameter,
                                 RangeDateMixin,
                                 TimezoneMixin):
    """Helper to build the payload sent to the API when retrieving
    social performance data.

    ..warning:: Only social accounts of some platforms can be analyzed. To
        check the available platforms, you can examine the
        ``AVAILABLE_PLATFORMs`` class attribute.

    Example:

        >>> from radarly.constants import PLATFORM
        >>> start = datetime(2018, 1, 1)
        >>> end = datetime(2018, 1, 31)
        >>> param = SocialPerformanceParameter() \\
                .timezone('Europe/Paris') \\
                .platform(PLATFORM.INSTAGRAM) \\
                .date_range(start, end)
    """
    AVAILABLE_PLATFORMS = [
        'instagram',
        'youtube',
        'twitter',
        'facebook',
        'linkedin',
        'sinaweibo'
    ]
    def platform(self, platform_value):
        """
        Args:
            platform_value (str): Platform of the social accounts on which the
                benchmark will be computed.
        """
        assert platform_value in self.AVAILABLE_PLATFORMS, \
            "The platform must be in {}".format(self.AVAILABLE_PLATFORMS)
        self['platform'] = platform_value
        return self

    def __call__(self):
        return urlencode(self)


class TopicParameter(Parameter,
                     StandardParameterMixin,
                     TimezoneMixin,
                     MetricsMixin,
                     LocaleMixin):
    """Helper to bulid the payload sent to the API when retrieving topic data.
    This helper contains all the methods to restrict the set of publications
    (methods defined in radarly.parameters.field.StandardParameterMixin).

    Example:

        >>> from radarly.constants import PLATFORM, METRIC
        >>> start = datetime(2018, 1, 1)
        >>> end = datetime(2018, 1, 31)
        >>> topic_param = TopicParameter() \\
                .locale('fr_FR') \\
                .metrics(METRIC.DOC, METRIC.IMPRESSION) \\
                .timezone('Europe/Paris') \\
                .focuses(include=[1, 2, 3, 4]) \\
                .publication_date(start, end) \\
                .platforms(PLATFORM.INSTAGRAM)
    """
    pass



"""
The distribution is a very useful feature in Radarly because it allows you
to get the distribution over time (by hour, day, week or month) of all your
publications.
"""

from .api import RadarlyApi


class Distribution(list):
    """
    List-like object storing the distribution over time of some metrics. Each item
    in the list is the statistics for one date. The ``Distribution`` object
    is compatible with ``pandas``.

    Example:

        >>> import pandas as pd
        >>> distribution
        Distribution(length=100)
        >>> df_distribution = pd.DataFrame(distribution)
        >>> df_distribution.set_index('date', inplace=True)
        >>> df_distribution.head()
                    doc  impression
        date
        2017-01-01  5678    55249194
        2017-01-02  6421    76283213
        2017-01-03  6489    69826116
        2017-01-04  7227    67470586
        2017-01-05  7646    68190331
    """
    def __init__(self, data):
        super().__init__()
        self.total = data['total']
        gen = (item for item in data['distribution'])
        for element in gen:
            temp = {'date': element['date']}
            temp.update(element['counts'])
            self.append(temp)

    def __repr__(self):
        return '<Distribution.length={}>'.format(len(self))

    @classmethod
    def fetch(cls, project_id, parameter, api=None):
        """Retrieve distribution data from the Radarly API.

        Args:
            project_id (int): identifier of your project
            parameter (DistributionParameter): parameter used to specify
                on which subset of publications the distribution must be
                computed. See ``DistributionParameter`` documentation to see
                how to build this object.
            api (RadarlyApi, optional): API used to make the
                request. If None, the default API will be used.
        Returns:
            Distribution: list-like object storing statistics by date.
        """
        api = api or RadarlyApi.get_default_api()
        url = api.router.distribution['fetch'].format(project_id=project_id)
        data = api.post(url, data=parameter)
        return cls(data['distribution'])



"""
Social account registred in Radarly
"""

from .model import SourceModel


class SocialAccount(SourceModel):
    """Object storing information retrieved from the API about a social account
    registred in your project.

    .. warning:: This object does not have the same keys depending on the
        platform of the social account.

    Here are some useful keys (also available as instance's attributes):

    Args:
        id (str): unique identifier of the social account
        platform (str): platform of the social account
        url (str): link to your social account
        ...
    """
    def __init__(self, data, platform):
        super().__init__()
        self.platform = platform
        self.add_data(data)

    @classmethod
    def _builder(cls, social_accounts):
        social_objects = []
        for platform in social_accounts:
            social_objects.extend([
                cls(account, platform) for account in social_accounts[platform]
            ])
        return social_objects

    def __repr__(self):
        return '<SocialAccount.id={}.platform={}>'.format(
            self['id'], self['platform']
        )



"""
`radarly-py` : a python client for Radarly API
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

:Author: Linkfluence
:Version: 1.0.12
:Licence: Apache-2.0

.. _official documentation: https://api.linkfluence.com
.. _quickstart tutorials: https://api.linkfluence.com/python/quickstart.html#use-cases

This package, developed by Linkfluence is a Python client in order to
ease interactions with the official REST API of Radarly products. ``radarly-py``
was designed in order to be compatible with standard Python objects but also
with :mod:`pandas` and with the JSON format. You want to transform a pivot table
into a :class:`pandas.DataFrame`? Make it simply with
``pandas.DataFrame(my_pivot_table)``. You want to serve your user's data with
a server: serialize your object with ``my_current_user.json()``.
Please read the `official documentation`_ of the project to know all you can
do with this Python client and feel free to report any bug or suggestion you
have to our developer team.


Some information about this package
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^
This package was built by trying to respect the architecture of Radarly. Most
objects which exists in the product have been translated in the
client.


About the objects
^^^^^^^^^^^^^^^^^
The ``radarly-py`` module uses mainly two kind of objects:

* object which inherits from the ``SourceModel`` and stores information.
  The objects ``Project``, ``User``, ``Tag`` or ``Dashboard`` are built on
  this model. These objects aim to give you access to all your qualitative data
  stored in our databases. All objects of this types share some methods as
  ``draw_structure`` in order to ease the comprehension of Radarly's data
  models. Each object has also some specific methods in order to navigate
  in the API (for example, you can get some distribution statistics directly
  from a ``Project`` object using the ``get_analytics`` method). Check the
  documentation of each object to get some help on how to use it.
* object storing quantitative data, which are based on :class:`list` or ``dict``.
  These objects, like ``Analytics``, ``PivotTable`` or ``Distribution``, are
  build to be easily parsed with ``pandas``.

Most objects defined in ``radarly-py`` are documented. You can get additional
information on how to use it with the `quickstart tutorials`_.
"""

from radarly.api import RadarlyApi


__title__ = 'radarly'
__version__ = '1.0.12'
__author__ = 'Linkfluence'
__url__ = 'https://api.linkfluence.com'
__licence__ = 'Apache-2.0'
__copyright__ = 'Copyright 2018 Linkfluence'



"""
The focuses in Radarly are used to store all the queries you defined
to capture publications.
"""

from .model import SourceModel
from .utils._internal import instance_builder


class Focus(SourceModel):
    """
    Dict-like object used to explore focus' data returned by the
    API.

    Args:
        id (int): unique identifier for the focus
        label (str): name of the focus (it corresponds to the name in
            Radarly)
        description (None or str): if available, describe the purpose of
            the focus
        created (datetime): creation date of the focus
        filter (dict): information about the query
    """
    def __init__(self, data):
        super().__init__(data)

    def __repr__(self):
        focus_id, label = getattr(self, 'id'), getattr(self, 'label')
        return '<Focus.id={}.label={}>'.format(focus_id, label)

    @classmethod
    def _builder(cls, focuses):
        return instance_builder(cls, focuses)



"""A field in parameter is a (key , value) pair in the dictionary sent as
payload to the API. For most fields, the value is checked in
order to be sure that the parameter is built properly. This way, bad
requests can be avoided."""

import pytz

from ..constants import (CATEGORY, GENDER, GEOTYPE, INTERVAL, MEDIA, METRIC,
                         ORDER, PLATFORM, TONE)
from ..utils.checker import (check_date, check_geocode, check_language,
                             check_list)


class AuthorMixin:
    """Field used to build all the author related parameters in the payload
    data"""
    def author_birth_date(self, start=None, end=None):
        """Restrict the publications to the right interval for author's
        birthdate.

        Args:
            start (datetime.datetime): Min birthdate of the author
            end (datetime.datetime): Max birthdate of the author
        """
        start = check_date(start)
        end = check_date(end)
        if start or end:
            self['birthDate'] = dict()
            if start: self['birthDate'].update(gt=start)
            if end: self['birthDate'].update(lt=end)
        return self

    def author_has_children(self, has_children):
        """Restricts to author that declares to have children.

        Args:
            has_children (bool): whether or note the author has children
        """
        assert isinstance(has_children, bool), \
            "'has_children' must be a boolean"
        self.update(hasChildren=has_children)
        return self

    def author_in_relationship(self, in_relationship):
        """Restricts to author that declares to be in a relationship

        Args:
            in_relationship (bool): whether or not the author must be in a
                relationship"""
        assert isinstance(in_relationship, bool), \
            "'in_relationship' must be a boolean"
        self.update(inRelationship=in_relationship)
        return self

    def author_verified(self, verified):
        """Restricts to author with certified accounts.

        Args:
            verified (bool): whether or not the author has a certified
                account
        """
        assert isinstance(verified, bool), 'verified must be a boolean'
        self.update(verified=verified)
        return self

    def author_id(self, *author):
        """Restric the set of publications to those which are created by
        specific authors.

        Args:
            *author (list[Dict[str, str]]): each author object is an object
                which support the __getitem__ method and must have both
                'id' and 'platform' keys.
        """
        author_ids = [
            '(user.{platform}.id:{uid})'.format(
                platform=author_id['platform'], uid=author_id['id']
            ) for author_id in author
        ]
        author_ids = ' OR '.join(author_ids)
        query = self.get('query', '')
        self['query'] = '{} AND ({})'.format(query, author_ids) if query \
            else '({})'.format(author_ids)
        return self


class CategoryMixin:
    """Field used to build the *category* parameter in the payload data"""
    def categories(self, *category_name):
        """List of the Radarly categories you want to restrict your search.

        Args:
            *category_name (str): list of categories. You can use the CATEGORY
                object of ``radarly.constants`` module to know the available
                category."""
        CATEGORY.check(category_name)
        if category_name: self.update(categories=list(category_name))
        return self


class ClusterMixin:
    def clusters(self, *cluster_ids):
        check_list(cluster_ids, str)
        self['stories'] = cluster_ids
        return self


class CorporaMixin:
    """Field used to build the *corpora* parameter in the payload data"""
    def corpora(self, *corpora_ids):
        """List of the Radarly registred corpora ids you want to search into"""
        check_list(corpora_ids, str)
        if corpora_ids: self.update(corpora=list(corpora_ids))
        return self


class DateMixin:
    """Field used to build all date related parameters in the payload data"""
    def creation_date(self, created_before=None, created_after=None):
        """Restricts using indexation date"""
        created_before = check_date(created_before)
        created_after = check_date(created_after)
        if created_before or created_after:
            self['date'] = dict()
            if created_before: self['date']['createdBefore'] = created_before
            if created_after: self['date'].update(createdAfter=created_after)
        return self

    def publication_date(self, start=None, end=None):
        """Restricts using publication date"""
        start = check_date(start)
        end = check_date(end)
        if start or end:
            if start: self.update({'from': start})
            if end: self.update(to=end)
        return self


class EmojiMixin:
    """Field used to build the *emojis* parameter in the payload data"""
    def emoji(self, charts=None, annotations=None):
        """Args:
            charts (list):
            annotations (list):
        """
        if charts or annotations:
            self['emoji'] = dict()
            if charts: self['emoji'].update(charts=charts)
            if annotations: self['emoji'].update(annotations=annotations)
        return self


class FctxMixin:
    """Field used to build the *fctx* parameter in the payload data"""
    def fctx(self, *fctx_list):
        self['fctx'] = list(fctx_list)
        return self


class FieldsMixin:
    """Field used to build the *fields* parameter in the payload data"""
    def fields(self, *fields_list):
        self['fields'] = list(fields_list)
        return self


class FlagMixin:
    """Field used to build the *flag* parameter in the payload data"""
    def flag(self, favorite=None, trash=None, retweet=None):
        """Enables the retrieving of favorites or non-favorites, of trashed
        publications or of retweets."""
        args = [f for f in [favorite, trash, retweet] if f]
        check_list(args, bool)
        self['flag'] = dict()
        self['flag'].update(favorite=favorite, trash=trash, rt=retweet)
        return self


class FocusMixin:
    """Field used to build the *focuses* parameter in the payload data"""
    def focuses(self, include=None, exclude=None):
        """List of the Radarly registred query ids you want to search into.

        Args:
            include (list): query ids to include
            exclude (list): query ids to exclude.
        """
        if include or exclude:
            self['focuses'] = []
        if include:
            check_list(include, int)
            self['focuses'].extend(
                [dict(id=query, include=True) for query in include]
            )
        if exclude:
            check_list(exclude, int)
            self['focuses'].extend(
                [dict(id=query, include=False) for query in exclude]
            )
        return self


class FollowerMixin:
    """Field used to build the *followers* parameter in the payload data"""
    def followers(self, minf=None, maxf=None):
        """Restricts to the min/max number of followers of a
        twitter/instagram/sinaweibo source."""
        if minf or maxf:
            self['followers'] = dict()
            if minf:
                assert isinstance(minf, int),  \
                    'Min values of followers must be an integer'
                self['followers'].update(gt=minf)
            if maxf:
                assert isinstance(maxf, int), \
                    'Max values of followers must be  an integer'
                self['followers'].update(lt=maxf)
        return self


class GenderMixin:
    """Field used to build the *genders* parameter in the payload data"""
    def genders(self, *gender_list):
        """Restricts to the given genders.

        Args:
            *gender_list (string): list of wanted gender.
        """
        GENDER.check(gender_list)
        self.update(genders=list(gender_list))
        return self


class GeoFilterMixin:
    def geofilter(self, polygon=None):
        self['geoFilter'] = dict()
        if polygon is None:
            polygon = [
                {"lon": -180, "lat": 81.01649},
                {"lon": 180, "lat": 81.01649},
                {"lon": 180, "lat": -58.94237},
                {"lon": -180, "lat": -58.94237},
                {"lon": -180, "lat": 81.01649}
            ]
        self['geoFilter']['filterPolygon'] = polygon
        return self


class GeoMixin:
    """Field used to build the *geo* parameter in the payload data"""
    def geo(self, gtype, glist):
        """List of items following geo.type (fr, gb,...); Restricts to the
        given countries, given by an ISO 3166-1 alpha-2

        Args:
            gtype (str): the type of geographical place (country or town)
            glist (list[str]): if ``gtype`` is country, you can pass a
                country name, its alpha-2 code, its alpha-3 or its official
                name.
        """
        GEOTYPE.check(gtype)
        if gtype == GEOTYPE.COUNTRY:
            glist = check_geocode(glist)
        self['geo'] = dict(type=gtype, list=glist)
        return self


class GeoTypeMixin:
    """Field used to specify the region type for geographical distribution"""
    def geo_type(self, gtype):
        """
        Args:
            gtype (str):
        """
        assert gtype in ['region', 'town'], \
            "gtype is an unknown option for the geographical type"
        self['geo_type'] = gtype
        return self


class IntervalMixin:
    """Field used to build the *interval* parameter in the payload data"""
    def interval(self, interval_value):
        INTERVAL.check(interval_value)
        self['interval'] = interval_value
        return self


class KeywordMixin:
    """Field used to build all keywords related parameter in the payload data"""
    def _builder_kw(self, name, *args):
        check_list(args, str)
        self.setdefault('keywords', {})
        self['keywords'][name] = list(args)
        return self

    def hashtags(self, *hashtags):
        """Restrics to the given hashtags.

        Args:
            *hashtags (list[str]):
        """
        return self._builder_kw('hashtags', *hashtags)

    def mentions(self, *mentions):
        """Restrics to the given @mentions

        Args:
            *mentions (list[str]):
        """
        return self._builder_kw('mentions', *mentions)

    def named_entities(self, *named_entities):
        """Restrics to the given named entities.

        Args:
            *named_entities (list[str]):
        """
        return self._builder_kw('namedEntities', *named_entities)

    def keywords(self, *keywords):
        """Restrics to the given keywords (manual or trigger tags).

        Args:
            *keywords (list[str]):
        """
        return self._builder_kw('keywords', *keywords)


class LanguageMixin:
    """Field used to build the *langauages* parameters in the payload data."""
    def languages(self, *language_list):
        """Restricts to the given languages, given by an ISO 639-1 code.

        Args:
            *language_list (string): list of all wanted languages. You can
                pass a language name, its alpha-2 code or its alpha-3 code.
        """
        language_list = check_language(language_list)
        self.update(languages=list(language_list))
        return self


class LocaleMixin:
    """Field used to build the *locale* parameter in the payload data"""
    def locale(self, locale_value):
        self['locale'] = locale_value
        return self


class MediaMixin:
    """Field used to build the *media* parameter in the payload data"""
    def media(self, *media_list):
        """Restricts to the given media types.

        Args:
            *media_list (string):
        """
        MEDIA.check(media_list)
        self['media'] = list(media_list)
        return self


class MetricsMixin:
    """Field used to build the *metrics* parameter in the payload data"""
    def metrics(self, *metrics_val):
        METRIC.check(metrics_val)
        self['metrics'] = list(metrics_val)
        return self


class PaginationMixin:
    """Field used to build all pagination related parameters in the payload
    data
    """
    def pagination(self, start=0, limit=25):
        assert isinstance(start, int), 'start index must be  an integer'
        assert isinstance(limit, int), 'limit must be  an integer'
        assert (start >= 0) and (limit > 0), 'Weird'
        self.update(start=start, limit=limit)
        return self

    def next_page(self):
        limit = self.get('limit', 25)
        start = self.get('start', -limit) + limit
        self.update(start=start, limit=limit)
        return self


class PlatformMixin:
    """Field used to build the *platforms* parameter in the payload data.
    You can check the available options in radarly.constants module."""
    def platforms(self, *platform_list):
        """Source platforms for publications in Radarly. The available
        platforms are available in the `platforms` object in the `constants`
        module of `radarly`.

        Args:
            *platform_list (string): list of all wanted platforms
        """
        PLATFORM.check(platform_list)
        self.update(platforms=list(platform_list))
        return self


class QueryMixin:
    """Field used to build the *query* parameter in the payload data"""
    def query(self, query):
        """A UTF-8 search query string of maximum 4K characters maximum,
        including operators. eg: “linkfluence AND radarly”. For additional
        information on how to build the query, you can check the official
        documentation.

        Args:
            query (string):
        """
        assert isinstance(query, str), 'Query shoud be a string.'
        self.update(query=query)
        return self


class RangeDateMixin:
    """Field used to build the range date parameter."""
    def date_range(self, start=None, end=None):
        start = check_date(start)
        end = check_date(end)
        if start or end:
            if start: self['from'] = start
            if end: self['to'] = end
        return self


class SortMixin:
    """Field used to build the sort related parameter in the payload data"""
    def sort_by(self, sort_by_filter):
        assert sort_by_filter in self.AVAILABLE_SORT_BY, \
            'Unknown sortBy parameter'
        self.update(sortBy=sort_by_filter)
        return self

    def sort_order(self, sort_order_filter):
        ORDER.check(sort_order_filter.lower())
        self.update(sortOrder=sort_order_filter)
        return self


class TagMixin:
    """Field used to build the *tags* parameter in the payload data"""
    def tags(self, user_tags=None, custom_fields=None):
        """List of Radarly registred influencers group tags or custom Fields
        values under format you want to restrict to."""
        if custom_fields:
            self.setdefault('tags', dict())
            self['tags']['customFields'] = custom_fields
        if user_tags:
            self.setdefault('tags', dict())
            self['tags']['userTags'] = user_tags
        return self


class TimezoneMixin:
    """Field used to build the *timezone* parameter in the payload data"""
    def timezone(self, time_zone):
        if time_zone not in pytz.all_timezones:
            raise ValueError('The timezone has not been recognized.')
        self['tz'] = time_zone
        return self


class ToneMixin:
    """Field used to build the *tones* parameter in the payload data"""
    def tones(self, *tone_list):
        """Restricts to the given tones. All available tones are given in
        the `tones` object of the `constants` module.

        Args:
            *tone_list (string):
        """
        TONE.check(tone_list)
        self.update(tones=list(tone_list))
        return self

class EmotionMixin:
    """Field used to build the *emotions* parameter in the payload data"""
    def emotions(self, *emotion_list):
        """Restricts to the given emotions. All available emotions are given in
        the `emotions` object of the `constants` module.

        Args:
            *emotion_list (string):
        """
        TONE.check(emotion_list)
        self.update(emotions=list(emotion_list))
        return self


class StandardParameterMixin(QueryMixin,
                             PlatformMixin,
                             LanguageMixin,
                             GenderMixin,
                             AuthorMixin,
                             ToneMixin,
                             MediaMixin,
                             KeywordMixin,
                             EmojiMixin,
                             FollowerMixin,
                             FlagMixin,
                             DateMixin,
                             GeoMixin,
                             FocusMixin,
                             CorporaMixin,
                             CategoryMixin,
                             TagMixin,
                             IntervalMixin):
    """Parameter which can be used in most of requests"""
    pass



"""
Setup script for ``radarly-py``
"""

from os.path import dirname, join
from setuptools import setup, find_packages


def readme():
    """Generate the README file of the package"""
    filepath = join(dirname(__file__), 'README.rst')
    with open(filepath) as file_readme:
        return file_readme.read()


setup(
    name='radarly-py',
    version='1.0.12',
    description="Python client for Radarly API",
    long_description=readme(),
    author='Linkfluence',
    author_email='help@linkfluence.com',
    packages=find_packages(),
    install_requires=[
        'requests',
        'lxml',
        'pytz',
        'python-dateutil',
        'pycountry',
    ],
    include_package_data=True,
    keywords='radarly linkfluence api',
    license='Apache-2.0',
    classifiers=[
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
    project_urls={
        'Documentation': 'https://api.linkfluence.com',
    },
    python_requires='>=3.4',
)



"""
In Radarly, you can monitor the performance of any of your own social accounts.
This module defines an object in order to retrieve information about that.
"""

from .api import RadarlyApi


class SocialPerformance(list):
    """List-like object storing distribution over time of some stats about a
    social account. This object is compatible with ``pandas``.

    >>> social_perf
    <SocialPerformance.uid=172489456479180.platform=facebook>
    >>> import pandas as pd
    >>> df_social_perf = pd.DataFrame(social_perf)
    >>> df_social_perf.head()
                checkins  comments  count  global-brand-root-id  likes  \\
    date
    2018-03-14         0       NaN    1.0                 71773    576
    2018-03-15         0       4.0    NaN                 71773    643
    2018-03-16         0       1.0    1.0                 71773    807
    2018-03-17         0       4.0    1.0                 71773    574
    2018-03-18         0       4.0    1.0                 71773    510
    ...
    """
    def __init__(self, data, platform):
        super().__init__()
        self.uid = data['uid']
        self.platform = platform
        source = []
        for item in data['stats']:
            temp = {'date': item['date']}
            temp.update(item['scores'])
            source.append(temp)
        self.extend(source)

    def __repr__(self):
        return '<SocialPerformance.uid={0.uid}.platform={0.platform}>'.format(
            self
        )

    @classmethod
    def fetch(cls, project_id, parameter, api=None):
        """Retrieve information about social account performance from the API.

        Args:
            project_id (int): identifier of the project
            parameter (SocialPerformanceParameter): object sent as
                payload to the API. See ``SocialPerformanceParameter`` to see
                how to build this object.
            api (RadarlyApi, optional): API used to make the
                request. If None, the default API will be used.
        Returns:
            SocialPerformance: list-like object compatible with ``pandas``
        """
        api = api or RadarlyApi.get_default_api()
        platform = parameter['platform']
        parameter = parameter()

        url = api.router.social_performance['fetch'].format(
            project_id=project_id
        )
        url = "{}?{}".format(url, parameter)
        data = api.get(url)

        return [cls(item, platform) for item in data if item['stats']]



"""
A tag is a label you can define in Radarly in order to sort publications.
"""

from .model import SourceModel


class Tag(SourceModel):
    """Dict-like object storing information about custom and user tags.

    Args:
        id (int): unique identifier of the tag
        type (str): *custom* or *user*
        label (str): name of the tags
        subtags (list[dict]): custom fields which can be created in Radarly
        ...
    """
    def __init__(self, data, tag_type):
        super().__init__()
        self.type = tag_type
        data['subtags'] = data.pop('values', [])
        super().add_data(data, None)

    def __repr__(self):
        label = getattr(self, 'label')
        return '<Tag.label={1}.type={0.type}>'.format(
            self, label
        )

    @classmethod
    def _builder(cls, tags):
        """Transform a dictionary of custom tags and user tags into a list of
        tags.

        Args:
            tags (dict): raw data about tags
        Returns:
            list[Tag]:
        """
        ans = [cls(custom_tag, 'custom') for custom_tag in tags['custom']]
        ans.extend([cls(custom_tag, 'user') for custom_tag in tags['user']])
        return ans



"""A corpus gathers several social accounts that can be monitored in Radarly.
"""

from .api import RadarlyApi
from .model import SourceModel


class Corpus(SourceModel):
    """Store information about a corpus.

    Args:
        created (datetime.datetime): creation date of the corpus
        id (int): unique identifier for the corpus
        name(str): name given to the corpus
        project_id (int): identifier of the project in which the corpus is
            registred.
        media_source (dict): dictionary which maps a platform with all the
            media sources available for this platform.
    """
    def __repr__(self):
        return '<Corpus.id={}.name={}>'.format(self['id'], self['name'])

    def __len__(self):
        """Get the number of media source registred in the corpus"""
        return sum([len(getattr(self, 'media_source')[platform])
                    for platform in getattr(self, 'media_source')])

    @classmethod
    def fetch_media(cls, project_id, corpora_id, api=None):
        """Retrieve all media sources linked to a corpus.

        Args:
            project_id (int): identifier of the project
            corpora_id (int): identifier of the corpus
            api (optional, RadarlyApi): API to use to perform the request.
                If ``None``, the default API will be used.
        """
        api = api or RadarlyApi.get_default_api()
        url = api.router.corpora['fetch_media'].format(
            project_id=project_id, corpora_id=corpora_id
        )
        return api.get(url)


class InfoCorpus(SourceModel):
    """Store small part of information about a corpus. In order to get the full
    object, you can use the ``expand`` method.

    Args:
        created (datetime.datetime): creation date of the corpus
        id (int): unique identifier for the corpus
        name(str): name given to the corpus
        project_id (int): identifier of the project in which the corpus is
            registred.
    """
    def __repr__(self):
        return '<InfoCorpus.id={}.name={}>'.format(self['id'], self['name'])

    @classmethod
    def _builder(cls, corpora):
        ans = []
        for corpus in corpora:
            del corpus['is_public']
            ans.append(cls(corpus))
        return ans

    def expand(self, api=None):
        """Get full data about a corpus (in particular to retrieve the media
        sources of the corpus.

        Args:
            api (optional, RadarlyApi): API object to use to perform the request. If
                ``None``, the default API will be used.
        Returns:
            Corpus: nearly the same object as ``InfoCorpus`` but with the
            ``media_source`` attribute in addition.
        """
        data_corpus = self.__dict__
        data_corpus['media_source'] = Corpus.fetch_media(self['project_id'],
                                                         self['id'],
                                                         api)
        return Corpus(data=data_corpus)



"""
Authentification class used by ``requests`` on each request. The authentication
system used by Radarly is based on a bearer token system.
"""

from reprlib import repr as truncate_repr

from requests.auth import AuthBase


class RadarlyAuth(AuthBase): # pylint: disable=R0903
    """Authorization object used in the request to access Radarly API.
    This object is automatically initialized by the ``RadarlyApi`` object
    using your access_token so you don't need to handle the authentication
    for each of your requests to the API. If you want to do some other requests
    using requests module without using the client python, you can use this
    object for authentication.

    >>> import requests
    >>> from radarly.auth import RadarlyAuth
    >>> auth = RadarlyAuth(<my_access_token>)
    >>> url_user = 'https://radarly.linkfluence.com/1.O/user.json'
    >>> requests.get(url_user, auth=auth)
    """
    def __init__(self, token):
        self.token = token

    def __call__(self, r):
        r.headers['Authorization'] = 'Bearer {.token}'.format(self)
        return r

    def __repr__(self):
        token = truncate_repr(self.token)
        return '<RadarlyAuth.token={}.type=bearer>'.format(token)



"""
Exception raised by the client.
"""

from requests.exceptions import HTTPError

from .utils._internal import _parse_error_response
from .utils.misc import dict_to_tsv


class NoInitializedApi(Exception):
    """Error raised if the module could not find an API to make a request"""
    def __str__(self):
        return (
            'No API has been initialized.'
        )


class AuthenticationError(Exception):
    """Error raised if (client_id, client_secret) doesn't match any
    credential in our database.
    """
    def __init__(self, error_type):
        super().__init__()
        self.type = error_type

    def __str__(self):
        if self.type == 'invalid_client':
            return 'The client_id or the client_secret is invalid.'
        elif self.type == 'invalid_scope':
            return ("The asked scopes are uncompatible with the scopes "
                    "linked to your account.")
        elif self.type == 'unsupported_grant_type':
            return "The grant_type is not currently supported."
        elif self.type == 'unauthorized_client':
            return "You don't have the permission to access the API."
        return 'Error: {}'.format(self.type)


class RadarlyHTTPError(HTTPError):
    """An HTTP error occured when querying the Radarly API"""
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.parsed_response = _parse_error_response(self.response)

    def __str__(self):
        content = self.parsed_response.get('error_message') \
            or self.response.text or '(no content)'
        data = {
            'Code': self.response.status_code,
            'Reason': self.response.reason,
            'URL': self.response.url,
            'Content': content,
        }
        if 400 <= data['Code'] < 500:
            data['Reason'] += ' (Client Side Error)'
        elif 500 <= data['Code'] < 600:
            data['Reason'] += ' (Servor Side Error)'
        return '{}\n{}'.format(
            'An error occured during the request.', dict_to_tsv(data)
        )


class RateReached(Exception):
    """Error raised when the user can not a request anymore."""
    def __init__(self, message):
        super().__init__(message)
        self.radarly_message = message


class PublicationUpdateFailed(Exception):
    """Error raised if the update of a publication's field failed.

    Args:
        not_updated_fields (list[str]): list of fields which have not been
            updated.
    """
    def __init__(self, fields):
        super().__init__()
        self.not_updated_fields = fields

    def __str__(self):
        return ("The field `{}` of the publication "
                "has not been updated.").format(
                    ", ".join(self.not_updated_fields)
                )



"""
An influencer in Radarly is an author of several publications who has a
more-or-less large audience. The influencer module of ``radarly-py`` defines
several methods and functions in order to help you to understand who are the
influencers in your project and consequently understand your audience.
"""

from .api import RadarlyApi
from .model import GeneratorModel, SourceModel
from .utils._internal import parse_struct_stat


class Influencer(SourceModel):
    """Dict-like object storing information about an influencer. The value of
    this object are available as value associated to a key, or as attribute of
    the instance. Here are some useful attributes:

    .. warning:: The structure of the ``Influencer`` object may change
        depending on the platform

    Args:
        id (str): identifier for the influencer
        platform (str): origin platform of the influencer
        screen_name (str): display name on the social_accounts
        permalink (str): link to the social_account
        followers_count (int): numbers of followers of the influencer
        count (int): number of documents published by the follower in your
            project.
        stats (dict): statitics about the influencers publications
    """
    def __init__(self, data, project_id):
        self.project_id = project_id
        super().__init__()
        translator = dict(
            stats=parse_struct_stat
        )
        if 'user' in data:
            super().add_data(data['user'], translator)
            del data['user']
        super().add_data(data, translator)

    def __repr__(self):
        influ_id, platform = getattr(self, 'id'), getattr(self, 'platform')
        return '<Influencer.id={}.platform={}>'.format(influ_id, platform)

    @classmethod
    def find(cls, project_id, influencer_id, platform, api=None):
        """Retrieve information about an influencer.

        Args:
            project_id (int): id of the project
            influencer_id (int): id of the influencer
            platform (str): platform of the influencer
            api (RadarlyApi, optional): API used to make the
                request. If None, the default API will be used.
        Returns:
            Influencer:
        """
        api = api or RadarlyApi.get_default_api()
        url = api.router.influencer['find'].format(project_id=project_id)
        params = dict(
            uid=influencer_id,
            platform=platform
        )
        res_data = api.get(url, params=params)
        return Influencer(res_data, project_id)

    @classmethod
    def fetch(cls, project_id, parameter, api=None):
        """Retrieve influencers list from a project.

        Args:
            project_id (int): id of the project
            parameter (InfluencerParameter): parameter sent as payload
                to the API. See ``InfluencerParameter`` to see how to build
                this object.
            api (RadarlyApi): API used to performed the request. If None, the
                default API will be used.
        Returns:
            list[Influencer]:
        """
        api = api or RadarlyApi.get_default_api()
        url = api.router.influencer['search'].format(project_id=project_id)
        data = api.post(url, data=parameter)
        return [cls(item, project_id) for item in data['users']]

    @classmethod
    def fetch_all(cls, project_id, parameter, api=None):
        """retrieve all influencers from a project.

        Args:
            project_id (int): identifier of a project
            parameter (InfluencerParameter): parameter sent as payload
                to the API. See ``InfluencerParameter`` to see how to build
                this object. This object must contain pagination's parameters.
            api (RadarlyApi): API used to performed the request. If None, the
                default API will be used.
        Returns:
            InfluencerGenerator:
        """
        return InfluencersGenerator(parameter,
                                    project_id=project_id, api=api)

    def get_metrics(self, api=None):
        """Retrieve metrics data about the influencer from the API.

        Returns:
            dict:
        """
        api = api or RadarlyApi.get_default_api()
        params = dict(
            platform=self['platform'],
            uid=self['id']
        )
        url = api.router.influencer['find'].format(project_id=self.project_id)
        metrics = api.get(url, params=params)['metrics']
        return metrics


class InfluencersGenerator(GeneratorModel):
    """Generator which yields all influencers matching some payload.

    Args:
        search_param (InfluencerParameter):
        project_id (int):
        api (RadarlyApi):
    Yields:
        Influencer:
    """
    def _fetch_items(self):
        """Get next range of influencers"""
        url = self._api.router.influencer['search'].format(project_id=self.project_id)
        res_data = self._api.post(url, data=self.search_param)
        self.total = 1000
        self._items = (
            Influencer(item, self.project_id) for item in res_data['users']
        )
        div = self.total // self.search_param['limit']
        reste = self.total % self.search_param['limit']
        self.total_page = div
        if reste != 0: self.total_page += 1
        self.search_param = self.search_param.next_page()

    def __repr__(self):
        return '<InfluencersGenerator.total={}.total_page={}>'.format(
            self.total, self.total_page
        )



"""
Model used in ``radarly-py``
"""

import copy
import json
from abc import ABC, abstractmethod

from pytz import timezone

from .api import RadarlyApi
from .utils.jsonparser import RadarlyEncoder
from .utils.path import dpath, draw_structure


class SourceModel:
    """Mixin to transform dictionary into an object where the keys of the
    dictionary are attributes of the instance"""
    _TRANSLATOR = dict(
        timezone=timezone
    )

    def __init__(self, data=None, translator=None):
        if data is None: data = dict()
        self.add_data(data, translator)

    def add_data(self, data, translator=None):
        """Add all (key, value) of data in the object

        Args:
            data (dict): data to transfer to the new object
            translator (dict, optional): translator to convert some value of the dict
        Returns:
            None:
        """
        translator = translator or dict()
        data = {
            key: translator.get(key, lambda x: x)(data[key])
            for key in data.keys()
        }
        self.__dict__.update(data)
        return None

    def __getitem__(self, path):
        """Quickly explore and find specific values in an object by following a
        path. This path must respect some format rules in order to be correctly
        understood by the function.

        * The path must begin with ``$``. If the path starts with ``$$``, the
        function will not raise an error if a key specified in the path is not
        found but will return None instead. If the ``path`` doesn't start with
        ``$``, the behaviour of the function is the same as the standard
        ``__getitem__`` function.
        * Each key of the path must be separated with a point ``.``
        * If you want to get only one element of a list, you can specify the index
        of the element, as in ``dashboards[3]``.
        * You can filter a list to only retrieve items matching a criteria. The
        criteria must be between parenthesis. The supported operator are: =, !=,
        <=, >=, >, <, in, notin

        **Example of valid paths**:

        * ``$.focuses[3].id``
        * ``$$.projects(id>50).limits.stop``
        * ``$.projects(DEMO in label).id``

        **Example of not valid paths**:

        * ``focuses[3].id`` (doesn't start with ``$``)

        Args:
            source (dict): object to explore
            path (str): path conformed to a JSON path
        Returns:
            object: object matching the dpath.
        """
        return dpath(self, path)

    def draw_structure(self, max_depth=5, show_type=True):
        """Draw the structure of an element

        Args:
            max_depth (int, optional): Default to 5.
            show_type (bool, optional): Default to True
        Returns:
            None:
        """
        draw_structure(element=(self.__class__.__name__, self),
                       level=0, max_depth=max_depth,
                       show_type=show_type, indent=2)
        return None

    def keys(self):
        """Returns a set of available attribute."""
        return {key for key in self.__dict__ if not key.startswith('_')}

    def json(self, **kwargs):
        """Serialize the object in order to have a dict object"""
        return json.dumps(self, cls=RadarlyEncoder, **kwargs)


class GeneratorModel(ABC):
    """Generator which yields all items matching some payload.

    Args:
        search_param (Parameter): parameter which should contains pagination
            parameters
        project_id (int, optional): identifier of the project
        api (RadarlyApi, optional): API used to perform request. If None,
            the default API will be used.
    Yields:
        object:
    """
    def __init__(self, search_param, project_id=None, api=None):
        self._api = api or RadarlyApi.get_default_api()
        self.project_id = project_id
        self.total = 0
        self.total_page = 0
        self.search_param = copy.deepcopy(search_param)
        self._items = None
        self.current_page = 1
        self._fetch_items()

    def __iter__(self):
        return self

    @abstractmethod
    def _fetch_items(self):
        """Get next range of items"""
        raise NotImplementedError(("In order to use this mixin, you must "
                                   "implement the _fetch_items method"))

    def __next__(self):
        try:
            returned_value = next(self._items)
        except StopIteration:
            self.current_page += 1
            if self.current_page > self.total_page:
                raise StopIteration
            self._fetch_items()
            returned_value = next(self._items)
        return returned_value



"""A milestone is a set of parameters used to add a label on a time period.
A milestone can then be used as a time's and timezone's filter."""

from .utils._internal import instance_builder
from .model import SourceModel


class Milestone(SourceModel):
    """This object stores all information about a milestone. Data store
    in the milestones can be used to configure the ``Parameter`` object.

    Args:
        id (str): unique identifier of the milestone
        name (str): name of the milestone
        description (str, optional): description of the milestone
        start_date (datetime.datetime): start date of the time period
            of the milestone.
        end_date (datetime.datetime): end date of the time period of
            the milestone
        interval (str): interval for the time period. Can be ``month``,
            ``week``, ``day`` or ``hour``.
        picture_id (str): ID of the custom picture for the milestone
        timezone (pytz.timezone): timezone of the milestone
        visibility (str): ``public`` or ``private``
    """
    @classmethod
    def _builder(cls, milestones):
        return instance_builder(cls, milestones)

    def __repr__(self):
        return '<Milestone.id={}.name={}>'.format(
            self['id'], self['name']
        )



"""
The Radarly API uses several constants especially in the search parameters.
In order to avoid you bad requests to the API, we have defined
several objects which represents all the constants which can be used to
parametrize the requests. These constants are used in ``radarly.parameters``
in order to check if the given parameters you enter are correct.
"""

from .utils.misc import load_data, dict_to_namedtuple, flat


__all__ = [
    'ANALYTICS_FIELD',
    'BY',
    'CATEGORY',
    'CLOUD_FIELD',
    'GENDER',
    'GEOTYPE',
    'INTERVAL',
    'MEDIA',
    'METRIC',
    'ORDER',
    'PLATFORM',
    'TONE',
    'EMOTION'
]


def _set_categories_data(cls):
    """Dynamically load the categories data in ``CATEGORY`` class

    Args:
        class in which the data must be injected
    Returns:
        None:
    """
    categories = load_data('assets/categories.json')
    for category_name in categories:
        dict_category = {
            item.replace('-', '_').upper(): '.'.join([category_name, item])
            for item in categories[category_name]
        }
        namedtuple = dict_to_namedtuple(category_name, dict_category)
        setattr(cls, category_name.upper(), namedtuple)
    return None


class CheckerMixin:
    """Mixin in order to check if the parameters are correct"""
    @classmethod
    def check(cls, elements):
        """Check that the given values for a specific key in payload data are
        right.

        Args:
            elements (list or string): if ``elements`` is a string, check if it
                is an available option for the key. If ``elements`` is a list
                or tuple, call the function on all item in the list.
        Raises:
            AssertionError: raised if one element isn't right.
        Returns:
            boolean: True if all is ok
        """
        if isinstance(elements, (list, tuple)):
            return all([cls.check(element) for element in elements])
        attributes = dict(cls.__dict__)
        for key in attributes.copy():
            if ((key.startswith('__') and key.endswith('__'))
                    or callable(attributes[key])
                    or isinstance(attributes[key], classmethod)):
                _ = attributes.pop(key)
        assert elements in attributes.values(), \
            "{} is an unknown option for the '{}' parameter".format(
                elements, cls.__name__
            )
        return True


class PLATFORM(CheckerMixin):
    """List of origin platforms for the publications in Radarly"""
    BLOG = 'blog'
    COMMENT = 'comment'
    DAILYMOTION = 'dailymotion'
    FACEBOOK = 'facebook'
    FORUM = 'forum'
    GPLUS = 'gplus'
    INSTAGRAM = 'instagram'
    LINKEDIN = 'linkedin'
    MEDIA = 'media'
    REVIEW = 'review'
    SINAWEIBO = 'sinaweibo'
    TWITTER = 'twitter'
    VKONTAKTE = 'vkontakte'
    WEBSITE = 'website'
    WECHAT = 'wechat'
    YOUKU = 'youku'
    YOUTUBE = 'youtube'


class TONE(CheckerMixin):
    """Tone of a publication given by our algorithms"""
    MIXED = 'mixed'
    NEGATIVE = 'negative'
    NEUTRAL = 'neutral'
    POSITIVE = 'positive'


class EMOTION(CheckerMixin):
    """Emotion of a publication given by our algorithms"""
    JOY = 'joy'
    LOVE = 'love'
    SURPRISE = 'surprise'
    FEAR = 'fear'
    ANGER = 'anger'
    DISGUST = 'disgust'
    FEAR = 'sadness'

class MEDIA(CheckerMixin):
    """Media's type of publications"""
    IMAGE = 'image'
    VIDEO = 'video'


class GEOTYPE(CheckerMixin):
    """Geographic area's type used for geographical distribution"""
    COUNTRY = 'country'
    TOWN = 'town'


class CATEGORY(CheckerMixin):
    """Categories defined by our algorithm which sorts posts by topics.

    .. note:: This class is initialized dynamically.
    """
    @classmethod
    def check(cls, elements):
        if isinstance(elements, (list, tuple)):
            return all([cls.check(element) for element in elements])
        attributes = dict(cls.__dict__)
        for key in attributes.copy():
            if ((key.startswith('__') and key.endswith('__'))
                    or callable(attributes[key])
                    or isinstance(attributes[key], classmethod)):
                _ = attributes.pop(key)
        attributes = flat([list(attributes[item]) for item in attributes])
        assert elements in attributes, \
            "{} is an unknown option for the '{}' parameter".format(
                elements, cls.__name__
            )
        return True

_set_categories_data(CATEGORY)


class GENDER(CheckerMixin):
    """Gender of people on the publication if available."""
    FEMALE = 'F'
    MALE = 'M'


class CLOUD_FIELD(CheckerMixin): # pylint: disable=C0103
    """Cloud fields which are available as parameters."""
    AFFECTS = 'affects'
    EMOJIS = 'emojis'
    HASHTAGS = 'hashtags'
    KEYWORDS = 'keywords'
    MENTIONS = 'mentions'
    NAMED_ENTITIES = 'namedEntities'


class ANALYTICS_FIELD: # pylint: disable=C0103
    """Kind of analytics you want to retrieve when you ask statistics
    about a query or set of queries."""
    CATEGORIES = 'categories'
    COUNTRIES = 'countries'
    DEMOGRAPHY = 'demography'
    DENSITY_PERSON = 'persons'
    FOCUSES = 'focuses'
    GENDERS = 'genders'
    IMAGE_GENDERS = 'imageGenders'
    LANGUAGES = 'languages'
    LOGOS = 'logos'
    OCCUPATIONS = 'occupations'
    OPERATING_SYSTEMS = 'operatingSystems'
    PLATFORMS = 'platforms'
    TONES = 'tones'
    EMOTIONS = 'emotions'


class METRIC(CheckerMixin):
    """Metrics on the application. Notice that other metrics then DOC need
    some computations so it can be slower."""
    DOC = 'doc'
    IMPRESSION = 'impression'
    REACH = 'reach'
    REPOST = 'repost'
    ENGAGEMENT = 'engagement'

class ORDER(CheckerMixin):
    """Parameter used in order to choose the sort order in a list returned
    by the API. This parameter shoud ideally be used with the sortBy parameter.
    """
    ASCENDANT = 'asc'
    DESCENDANT = 'desc'

class BY(CheckerMixin):
    """Parameter used to sort the results returned by the API"""
    DATE = 'date'
    ENGAGEMENT = 'radar.engagement'
    IMPRESSION = 'radar.impression'
    INFLUENCER_IMPRESSION = 'impressions'
    INFLUENCER_POST = 'post'
    INFLUENCER_REACH = 'reach'
    RANDOM = 'random'
    RATING = 'radar.rating'
    REACH = 'radar.reach'
    VIRALITY = 'radar.virality'
    VOLUMETRY = 'volumetry'

class INTERVAL(CheckerMixin):
    """Interval to use in order to group results when a temporal distribution
    is returned by the API."""
    DAY = 'day'
    HOUR = 'hour'
    MONTH = 'month'
    YEAR = 'year'
